import typing

from . import _helpers
from . import _intel
from . import _ansi
from . import _source


__all__ = ('Terminate', 'Handle')


class Terminate(Exception):

    """
    Signals the invokation loop to stop.
    """


_type_Handle_start_invoke = typing.Callable[[str, _ansi._type_parse_return], None]

    
class Handle:

    """
    Encapsulates the invokation loop logic.
    """

    __slots__ = ('_intel',)

    def __init__(self, 
                 intel: _intel.Intel):

        self._intel = intel

    def _start(self, invoke):

        source = _source.Source(invoke, self._intel)

        try:
            source.listen()
        except Terminate:
            pass

    def start(self,
              invoke: _type_Handle_start_invoke):
        
        """
        Start the invokation loop.

        This will use :meth:`.Source.listen` until :exc:`Terminate` is raised.

        :param invoke:
            Used as :paramref:`.Source.callback` callback.
        """

        self._start(invoke)



"""
Classes for printing data that can change over time.
"""

import typing
import abc
import functools
import time
import math
import threading
import collections

from . import _core
from . import _helpers
from . import _system
from . import _colors
from . import _theme


__all__ = ('Graphic', 'BaseProgress', 'SpinProgress', 'LineProgress')


class Graphic(abc.ABC):

    """
    Base for graphics.

    :param inline:
        Whether ``\\n`` should **not** be used upon closing.
    :param throttle:
        Seconds for printing time throttling.
    :param epilogue:
        Shown upon closing instead of the completed graphic.
    """

    _lock = threading.RLock()

    _top_point = (0, 0)
    _cur_count = 0

    def __init__(self, 
                 inline   : bool = False, 
                 throttle : float = 0.1,
                 epilogue : typing.Union[str, typing.Callable[[], str], None] = None):

        self._inline = inline

        self._throttle = throttle
        self._time_last = 0

        self._epilogue = epilogue

        io = _system.io
        cursor = _system.cursor
        render = _core.Render(cursor, io)
        screen = _core.Screen(render, cursor)

        self._io = io
        self._cursor = cursor
        self._render = render
        self._screen = screen

        self._cur_index = None
        self._cur_point = None

    @abc.abstractmethod
    def _get(self):

        return NotImplemented
    
    @property
    def _final(self):

        return self._cur_point == self._top_point

    def _sketch(self, close):

        value = self._epilogue
        if close and not value is None:
            if callable(value):
                value = value()
        else:
            value = self._get()

        lines = _helpers.split_lines(value)
        point = None

        return (lines, point)
    
    @_helpers.ctxmethod(lambda self: self._lock)
    def _print(self, check, *, close = False):

        time_next = time.time()

        if check and time_next - self._time_last < self._throttle:
            return
        
        self._time_last = time_next

        sketch = functools.partial(self._sketch, close)

        self._screen.print(sketch, True, clean = False)

        if not close:
            self._cur_point = point = self._cursor.locate()
            self.__class__._top_point = max(self._top_point, point)

    @_helpers.ctxmethod(lambda self: self._lock)
    def _start(self):

        if not self._cur_index is None:
            return

        self._cur_index = self._cur_count

        self.__class__._cur_count += 1

        if self._cur_index:
            self._render.draw([[], []], learn = False, clean = False)

        self._print(False)

    def start(self):

        """
        Start the graphic. Calling this is required before further use.
        """

        self._start()

    @_helpers.ctxmethod(lambda self: self._lock)
    def _close(self):

        self._print(False, close = True)

        self.__class__._cur_count -= 1

        if self._cur_count:
            return
        
        if not self._final:
            self._cursor.move(*self._top_point)

        if not self._inline:
            sketch = lambda: ([[], []], None)
            self._screen.print(sketch, False, learn = False)

    def close(self):

        """
        Close the graphic. Calling this is required after complete use.
        """

        self._close()
 
    def __enter__(self, *args):

        self._cursor.hidden.enter()

        self._start()

        return self

    def __exit__(self, *args):

        self._close()

        self._cursor.hidden.leave()


class BaseProgress(Graphic):

    """
    Base for progress graphics.

    :param frequency:
       The amount to wait between automatically adding 1.
    """

    def __init__(self, 
                 *args, 
                 frequency: float = None,
                 **kwargs):
        
        self._frequency = frequency

        self._cycle_actual = 0
        self._cycle_memory = 0

        super().__init__(*args, **kwargs)

    def _print(self, check, *args, **kwargs):

        size = self._cycle_memory

        if check and not size > 0:
            return
        
        self._cycle_actual += size
        self._cycle_memory = 0

        super()._print(check, *args, **kwargs)

    def _start_auto(self):

        while not self._frequency_rolling is None:
            self._add(1)
            time.sleep(self._frequency_rolling)

    def _start(self):

        super()._start()

        self._frequency_rolling = self._frequency

        self._thread = thread = threading.Thread(target = self._start_auto, daemon = True)
        
        thread.start()

    def _close(self):

        self._frequency_rolling = None

        super()._close()

    def _set(self, step):

        self._cycle_memory = max(0, step)

        self._print(True)

    def set(self, step):

        """
        Set the progress to a specific step.
        """

        self._set(step)

    def _add(self, size):

        step = self._cycle_memory + size
        
        self._set(step)

    def add(self, size):

        """
        Add to the current progress step.
        """

        self._add(size)


class SpinProgress(BaseProgress):

    """
    Progress "spinner" with undeterminable ending condition.

    :param phases:
        The phases to cycle over when adding ``1`` progress.
    :param prefix:
        Prepended to the phase.
    :param suffix:
        Appended to the phase.
    
    .. code-block:: python

        with survey.graphics.SpinProgress(prefix = 'Loading ', epilogue = 'Done!') as progress:
            for _ in range(10):
                progress.add(1)
                time.sleep(.35)

    |theme| :code:`graphics.SpinProgress`.

    .. image:: /_static/images/graphics.spinprogress-1.gif
    """

    @_theme.add('graphics.SpinProgress')
    def __init__(self, 
                 *args,
                 phases: typing.List[str] = ('.', 'o', 'O', '@', '*'),
                 prefix: str       = ' ',
                 suffix: str       = '',
                 **kwargs):

        self._prefix = prefix
        self._suffix = suffix

        self._phases = phases

        super().__init__(*args, **kwargs)

    def _get(self):

        cycle = self._cycle_actual % len(self._phases)

        value = (
              self._prefix
            + self._phases[cycle] 
            + self._suffix
        )

        return value
    

class LineProgress(BaseProgress):

    """
    Progress line with a clear start and end.

    :param total:
        The maximum progress value.
    :param width:
        The length of the line.
    :param phases:
        The fractional phases possible. The last one is used for a full bar. 
        Multiple full bars and at most 1 fractional make up the whole line.
    :param empty:
        Used for filling in the empty space.
    :param prefix_wall:
        Prepended to the line.
    :param suffix_wall:
        Appended to the line.
    :param prefix:
        Prepended to the final value.
    :param suffix:
        Appended to the final value.
    :param percentage_zfill_value:
        Used for padding the percentage.
    :param percentage_zfill_color:
        Used for painting the percentage padding.
    :param percentage_zfill_count:
        The maximum amount of percentage padding.
    :param throughput_zfill_value:
        Used for padding the throughput.
    :param throughput_zfill_color:
        Used for painting the throughput padding.
    :param throughput_zfill_count:
        The maximum amount of throughput padding.

    The final structure is as follows:

    .. code-block::

        (prefix)(percentage + ' ')(prefix_wall)(phases[-1]*?)(phases[?])(empty*?)(suffix_wall)(' ' + throughput)(suffix)

    .. code-block:: python

        with survey.graphics.LineProgress(100, prefix = 'Loading ') as progress:
            for _ in range(100):
                progress.add(1)
                time.sleep(0.035)

    |theme| :code:`graphics.LineProgress`.

    .. image:: /_static/images/graphics.lineprogress-1.gif
    """

    @_theme.add('graphics.LineProgress')
    def __init__(self, 
                 total                 : int, 
                 *args,
                 width                 : int       = 50,
                 phases                : typing.List[str] = ('-',), 
                 empty                 : str       = ' ',
                 prefix_wall           : str       = '[', 
                 suffix_wall           : str       = ']',
                 prefix                : str       = '',
                 suffix                : str       = '',
                 percentage_zfill_value: str       = '0',
                 percentage_zfill_color: str       = _colors.basic('black'),
                 percentage_zfill_count: str       = 3,
                 percentage_template   : str       = '{value}%',
                 throughput_zfill_value: str       = '0',
                 throughput_zfill_color: str       = _colors.basic('black'),
                 throughput_zfill_count: str       = 3,
                 throughput_template   : str       = '{value}/s',
                 **kwargs):

        self._prefix = prefix
        self._suffix = suffix

        self._total = total

        self._width = width

        self._empty = empty
        self._phases = phases
        self._prefix_wall = prefix_wall
        self._suffix_wall = suffix_wall

        self._percentage_template = percentage_template
        self._percentage_zfill_value = percentage_zfill_value
        self._percentage_zfill_color = percentage_zfill_color
        self._percentage_zfill_count = percentage_zfill_count
        
        self._throughput_template = throughput_template
        self._throughput_zfill_value = throughput_zfill_value
        self._throughput_zfill_color = throughput_zfill_color
        self._throughput_zfill_count = throughput_zfill_count
        self._throughput_deque = collections.deque(maxlen = 10)

        self._width = width

        super().__init__(*args, **kwargs)

    def _set(self, step):

        self._throughput_deque.append((time.time(), step))

        super()._set(step)

    def _get(self):

        size_limit = self._cursor.measure()[1]
        size_limit -= self._cursor.locate()[1]

        size_given = self._width

        if not size_given is None:
            size_limit = min(size_given, size_limit)

        size_ratio = min(self._total, self._cycle_actual) / self._total

        percentage = ''
        percentage_template = self._percentage_template
        if not percentage is None:
            percentage_value = str(math.ceil(size_ratio * 100))
            percentage_zfill_count = self._percentage_zfill_count - len(percentage_value)
            percentage_zfill_value = self._percentage_zfill_value * percentage_zfill_count
            percentage_zfill_color = self._percentage_zfill_color
            if not percentage_zfill_color is None:
                percentage_zfill_value = _helpers.paint_text(percentage_zfill_color, percentage_zfill_value)
            percentage_value = percentage_zfill_value + percentage_value
            percentage = percentage_template.format(value = percentage_value)
            percentage = percentage + ' '

        full_size_limit = size_limit
        full_size_ratio = size_ratio
        full_size_precise = full_size_limit * full_size_ratio
        full_size = round(full_size_precise)

        half_size_limit = len(self._phases)
        half_size_ratio = full_size_precise - full_size
        half_size_precise = half_size_limit * half_size_ratio
        half_size = math.floor(half_size_precise)
        half_index = half_size % half_size_limit

        add_size = full_size_limit - full_size

        throughput = ''
        throughput_template = self._throughput_template
        if not throughput_template is None:
            try:
                throughput_times, throughput_steps = zip(*self._throughput_deque)
                throughput_period = max(throughput_times) - min(throughput_times)
                throughput_amount = sum(throughput_steps)
                throughput_ratio = throughput_amount / throughput_period
            except (ValueError, ZeroDivisionError):
                throughput_value = ''
            else:
                throughput_value = str(math.ceil(throughput_ratio * 100))
            throughput_zfill_count = self._throughput_zfill_count - len(throughput_value)
            throughput_zfill_value = self._throughput_zfill_value * throughput_zfill_count
            throughput_zfill_color = self._throughput_zfill_color
            if not throughput_zfill_color is None:
                throughput_zfill_value = _helpers.paint_text(throughput_zfill_color, throughput_zfill_value)
            throughput_value = throughput_zfill_value + throughput_value
            throughput = throughput_template.format(value = throughput_value)
            throughput = ' ' + throughput

        line = (
              self._prefix
            + percentage
            + self._prefix_wall 
            + self._phases[- 1] * full_size
            + self._phases[half_index] 
            + self._empty * add_size
            + self._suffix_wall
            + throughput
            + self._suffix
        )

        return line



from . import _helpers


__all__ = ('BaseIO', 'ModeIO', 'StreamIO')


class BaseIO:

    __slots__ = ('_i', '_o')

    def __init__(self, i, o):

        self._i = i
        self._o = o

    def _send(self, text):

        raise NotImplementedError()

    def send(self, 
             text: str):

        """
        Send text to the output.

        :param text:
            The text to send.
        """

        self._send(text)

    def _recv(self):

        raise NotImplementedError()

    def recv(self) -> str:

        """
        Wait and receive text from the input.
        """

        text = self._recv()

        return text


class ModeIO(BaseIO):

    __slots__ = ('_atomic', '_i_save', '_i_mode')

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self._atomic = _helpers.Atomic(self._start, self._stop)
        
    def _init(self):

        raise NotImplementedError()

    def _set(self, mode):

        raise NotImplementedError()

    def _wait(self, mode, switch):

        raise NotImplementedError()

    def wait(self, 
             switch: bool):

        """
        Change the receiving mode waiting behavior.
        """

        mode = self._wait(self._i_mode, switch)

        self._set(mode)

    def _start(self):

        (self._i_save, mode) = self._init()

        mode = self._wait(mode, True)

        self._set(mode)

        self._i_mode = mode

    def _stop(self):

        self._set(self._i_save)

        self._i_save = None

    @property
    def __enter__(self):

        return self._atomic.__enter__

    @property
    def __exit__(self):

        return self._atomic.__exit__


class StreamIO(ModeIO):

    __slots__ = ()

    def _send(self, value):

        self._o.write(value)
        self._o.flush()

    def _recv(self):

        text = self._i.read1()

        return text

    def __iter__(self):

        while not self._i.closed:
            yield self._recv()




"""
Functions that can be used for changing the way data appears upon printing.

Specifically, they are meant to be passed to :paramref:`~.visuals.Visual.funnel_enter` or :paramref:`~.visuals.Visual.funnel_leave`.
All ``text_`` funnels can be used for the latter, while anythign else should be used former the former.


"""

import typing
import enum
import operator
import itertools
import collections
import builtins
import functools

from . import _helpers
from . import _colors


__all__ = (
    'text_replace', 'JustType', 'text_min_horizontal', 'text_min_vertical', 'text_max_horizontal', 'text_max_vertical',
    'text_max_dynamic', 'text_bloat_horizontal', 'text_bloat_vertical', 'text_block', 'text_paint', 'mesh_delegate',
    'mesh_delimit', 'mesh_focal', 'mesh_light', 'mesh_point', 'mesh_max', 'mesh_min', 'mesh_grid_fill', 'mesh_grid', 
    'mesh_head', 'mesh_flip'
)


def _call_direct(function):

    def direct(*args, **kwargs):
        (*args, lines, point) = args
        return function(*args, **kwargs)(lines, point)
    
    function.call = direct

    return function


def _text_replace(rune,
                  lines, point):

    get_rune = rune if callable(rune) else lambda *args: rune
    
    for line in lines:
        for index in range(len(line)):
            line[index] = get_rune(line[index])


@_call_direct
def text_replace(rune: typing.Union[str, typing.Callable[[str], str]]):

    """
    Replace every line element.

    :param rune:
        The rune to replace with, or a function that takes the current rune and returns the replacement.

    .. code-block::

        'The quic k brown fox'
        'jumps ov|er the lazy dog'
        'and fall s on its face'

        >>> text_replace('*')

        '******** ***********'
        '********|***************'
        '******** *************'
    """

    return functools.partial(_text_replace, rune)


class JustType(str, enum.Enum):

    """
    JustType()

    Denotes the type of alignment.
    """
    
    #: left or top
    start = 'start'
    #: right or bottom
    end = 'end'
    #: middle
    center = 'center'


def _text_min_horizontal(just, size, rune, 
                         lines, point):

    max_line_size = builtins.max(map(len, lines)) if size is None else size

    for (rol_y, rol_line) in enumerate(lines):
        padding_size = max_line_size - len(rol_line)
        if not padding_size > 0:
            continue
        if just == JustType.start:
            padding_size_left = 0
            padding_size_right = padding_size
        elif just == JustType.end:
            padding_size_left = padding_size
            padding_size_right = 0
        elif just == JustType.center:
            padding_size_left = padding_size // 2
            padding_size_right = padding_size - padding_size_left
        rol_line[0:0] = (rune,) * padding_size_left
        rol_line.extend((rune,) * padding_size_right)
        if rol_y == point[0]:
            point[1] += padding_size_left


@_call_direct
def text_min_horizontal(just: JustType, 
                        size: int, 
                        rune: str):
    
    """
    Ensure each line is at least of a certain length, aligned accordingly using a rune.
    
    :param just:
        Denotes justification to align with.
    :param size:
        The minimum length of each line.
    :param rune:
        The rune used to fill empty space.

    .. code-block::

        'The quic k brown fox'
        'jumps ov|er the lazy dog'
        'and fall s on its face'

        >>> text_min_horizontal(JustType.center, 25, ' ')

        '   The qu ick brown fox   '
        ' jumps ov|er the lazy dog '
        '  and fal ls on its face  '
    """

    return functools.partial(_text_min_horizontal, just, size, rune)


def _text_min_vertical(just, size, rune,
                       lines, point):

    lines_size = len(lines) if size is None else size

    padding_size = lines_size - len(lines)

    if not padding_size > 0:
        return

    if just == JustType.start:
        padding_size_top = 0
        padding_size_bottom = padding_size
    elif just == JustType.end:
        padding_size_top = padding_size
        padding_size_bottom = 0
    elif just == JustType.center:
        padding_size_top = padding_size // 2
        padding_size_bottom = padding_size - padding_size_top

    max_line_size = builtins.max(map(len, lines))

    lines[0:0] = ([rune] * max_line_size for _ in range(padding_size_top))
    lines.extend(([rune] * max_line_size for _ in range(padding_size_bottom)))

    point[0] += padding_size_top


@_call_direct
def text_min_vertical(just: JustType, 
                      size: int, 
                      rune: str):
    
    """
    Ensure there is at least a certain amount of lines, aligned accordingly using a rune.

    :param just:
        Denotes justification to align with.
    :param size:
        The minimum amount lines.
    :param rune:
        The rune used to create lines to fill empty space.

    .. code-block::

        'The quic k brown fox'
        'jumps ov|er the lazy dog'
        'and fall s on its face'

        >>> text_min_vertical(JustType.end, 5, '~')

        '~~~~~~~~ ~~~~~~~~~~~~~~~'
        '~~~~~~~~ ~~~~~~~~~~~~~~~'
        'The quic k brown fox    '
        'jumps ov|er the lazy dog'
        'and fall s on its face  '
    """

    return functools.partial(_text_min_vertical, just, size, rune)


def _text_max_horizontal(size,
                         lines, point):

    cur_y = point[0]
    cur_x = point[1]
    cur_line = lines[cur_y]

    min_max_x = size
    max_max_x = len(cur_line)
    min_min_x = 0
    max_min_x = max_max_x - min_max_x

    cut_x = size // 2

    min_x = builtins.max(min_min_x, builtins.min(max_min_x, cur_x - cut_x))
    max_x = builtins.max(min_max_x, builtins.min(max_max_x, min_x + size))

    for rol_line in lines:
        rol_line[max_x:] = rol_line[:min_x] = ()

    point[1] = cur_x - min_x


@_call_direct
def text_max_horizontal(size: int):
    
    """
    Ensure each line is at most a certain size, ensuring the cursor is visible.

    :param size:
        The maximum length of each line.

    .. code-block::

        'The quic k brown fox'
        'jumps ov|er the lazy dog'
        'and fall s on its face'

        >>> text_max_horizontal(10)

        ' quic k bro'
        'ps ov|er th'
        ' fall s on '
    """

    return functools.partial(_text_max_horizontal, size)


def _text_max_vertical(size,
                       lines, point):

    cur_y = point[0]

    min_max_y = size
    max_max_y = len(lines)
    min_min_y = 0
    max_min_y = max_max_y - min_max_y

    cut_y = size // 2

    min_y = builtins.max(min_min_y, builtins.min(max_min_y, cur_y - cut_y))
    max_y = builtins.max(min_max_y, builtins.min(max_max_y, min_y + size))

    lines[max_y:] = lines[:min_y] = ()

    point[0] = cur_y - min_y


@_call_direct
def text_max_vertical(size: int):
    
    """
    Ensure there is at most a certain amount of lines, ensuring the cursor is visible.

    :param size:
        The maximum amount of lines.

    .. code-block::

        'The quic k brown fox'
        'jumps ov|er the lazy dog'
        'and fall s on its face'

        >>> text_max_vertical(1)

        'jumps ov|er the lazy dog'
    """

    return functools.partial(_text_max_vertical, size)


def _text_max_dynamic(get,
                      lines, point):
    
    functions = (_text_max_vertical, _text_max_horizontal)
    
    sizes = get()

    for (function, size) in zip(functions, sizes):
        function(size, lines, point)


@_call_direct
def text_max_dynamic(get: typing.Callable[[], typing.Tuple[int, int]]):
    
    """
    A combination of :func:`.text_max_horizontal` and :func:`.text_max_vertical`.
    
    :param get:
        Should return an iterable with each function's size on their axis index.
    """

    return functools.partial(_text_max_dynamic, get)


def _text_bloat_horizontal(just, size, runes,
                           lines, point):

    push = builtins.max(map(len, lines))

    for (index, rune) in enumerate(runes):
        _text_min_horizontal(just, size + push + index, rune, lines, point)


@_call_direct
def text_bloat_horizontal(just: JustType, 
                          size: int, 
                          rune: str):
    
    """
    Increase each line's length to a certain size, aligned by an axis using a rune.

    :param just:
        Denotes justification to align with.
    :param size:
        The additional length of the longest line, applied as minimum to all.
    :param rune:
        The rune used to fill empty space.

    .. code-block::

        'The quic k brown fox'
        'jumps ov|er the lazy dog'
        'and fall s on its face'

        >>> text_bloat_horizontal(JustType.start, 2, '%')

        'The quic k brown fox%%%%%%'
        'jumps ov|er the lazy dog%%'
        'and fall s on its face%%%%'
    """

    return functools.partial(_text_bloat_horizontal, just, size, rune)


def _text_bloat_vertical(just, size, runes,
                         lines, point):

    push = len(lines)

    for (index, rune) in enumerate(runes):
        _text_min_vertical(just, size + push + index, rune, lines, point)


@_call_direct
def text_bloat_vertical(just: JustType, 
                        size: int, 
                        rune: str):
    
    """
    Increase the amount of lines lines to a certain size, aligned by an axis using a rune.

    :param just:
        Denotes justification to align with.
    :param size:
        The additional amount lines.
    :param rune:
        The rune used to create lines to fill empty space.

    .. code-block::

        'The quic k brown fox'
        'jumps ov|er the lazy dog'
        'and fall s on its face'

        >>> text_bloat_vertical(JustType.center, 2, '%')

        '%%%%%%%% %%%%%%%%%%%%%%%'
        'The quic k brown fox'
        'jumps ov|er the lazy dog'
        'and fall s on its face'
        '%%%%%%%% %%%%%%%%%%%%%%%'
    """

    return functools.partial(_text_bloat_vertical, just, size, rune)


def _text_block(rune_top, rune_bottom, rune_left, rune_right,
                rune_top_left, rune_top_right, rune_bottom_left, rune_bottom_right,
                lines, point):

    if rune_top:
        line = lines[0]
        size = len(line)
        line[:] = (rune_top,) * size

    if rune_bottom:
        line = lines[- 1]
        size = len(line)
        line[:] = (rune_top,) * size

    if rune_left:
        for line in lines:
            line[0] = rune_left
    
    if rune_right:
        for line in lines:
            line[- 1] = rune_right

    if rune_top_left:
        lines[0][0] = rune_top_left

    if rune_top_right:
        lines[0][- 1] = rune_top_right

    if rune_bottom_left:
        lines[- 1][0] = rune_bottom_left

    if rune_bottom_right:
        lines[- 1][ -1] = rune_bottom_right


@_call_direct
def text_block(rune_top: str, 
               rune_bottom: str, 
               rune_left: str, 
               rune_right: str,
               rune_top_left: str, 
               rune_top_right: str, 
               rune_bottom_left: str, 
               rune_bottom_right: str):
    
    """
    Surround the lines by overwriting with the respective runes. Any of them can be :code:`None` to ignore.

    :param rune_top:
        Used on the first line.
    :param rune_bottom: 
        Used on the last line.
    :param rune_left:
        Used on the first rune of each line.
    :param rune_right:
        Used on the last rune of each line.
    :param rune_top_left:
        Used on the first rune of the first line.
    :param rune_top_right:
        Used on the last rune of the first line.
    :param rune_bottom_left:
        Used on the first rune of the last line.
    :param rune_bottom_right:
        Used on the last rune of the last line.

    .. code-block::

        'The quic k brown fox'
        'jumps ov|er the lazy dog'
        'and fall s on its face'

        >>> text_block('-', '-', '|', '|', '/', '\\', '\\', '/')

        '/----------------------\\'
        '|umps ov|er the lazy do|'
        '\----------------------/'

    Use :code:`text_bloat_vertical(JustType.center, 2, ' ')` and :code:`text_bloat_horizontal(JustType.center, 2, ' ')` before this to avoid content overwrite.
    """

    return functools.partial(
        _text_bloat_vertical, rune_top, rune_bottom, rune_left, rune_right,
        rune_top_left, rune_top_right, rune_bottom_left, rune_bottom_right
    )

  
def _text_paint(color, lines, point):

    if callable(color):
        color = color()

    if color is None:
        return

    _helpers.paint_lines(color, lines)


@_call_direct
def text_paint(color: str):

    """
    Paint each rune of each line.

    :param color:
        The color to paint with.

    .. code-block::

        'The quic k brown fox'
        'jumps ov|er the lazy dog'
        'and fall s on its face'

        >>> text_paint('\x1b[36m')

        # The same, but in light blue.
    """

    return functools.partial(_text_paint, color)


def _mesh_delegate(check, aware, function, 
                   tiles, point):

    for (spot, tile) in tiles.items():
        if check and not check(spot, tile):
            continue
        args = [*tile]
        if aware:
            args.insert(0, spot)
        function(*args)


@_call_direct
def mesh_delegate(function: typing.Callable[[typing.List[typing.List[str]], typing.List[int]], None], 
                  check   : typing.Callable[[typing.Tuple[int, int], typing.List[typing.List[typing.List[str]]], typing.List[int]], bool] = None, 
                  aware   : bool = False):
    
    """
    Call a function on each tile of the mesh.

    :param function:
        The function called.
    :param check:
        Used as :code:`(spot, tile) -> bool` to denote whether the specific tile gets ignored.
    :param aware:
        Whether to prepend the tile's ``spot`` to the function's arguments.
    """

    return functools.partial(_mesh_delegate, check, aware, function)


def _mesh_delimit(axis, rune,
                  tiles, point):
    
    all_a = (spot[axis] for spot in tiles)
    max_a = max(all_a)

    def check(spot, tile):
        return spot[axis] < max_a
    
    bloat = _text_bloat_horizontal if axis else _text_bloat_vertical

    function = functools.partial(bloat, JustType.start, 1, rune)

    _mesh_delegate(check, False, function, tiles, point)


@_call_direct
def mesh_delimit(axis: int, 
                 rune: str):
    
    """
    Insert text between tiles.

    :param axis:
        The axis considered for the operation.
    :param rune:
        The text to insert.
    """

    return functools.partial(_mesh_delimit, axis, rune)


def _mesh_focal(function,
                tiles, point):
    
    main_spot = tuple(point)

    def delegate_function(spot, *tile):
        current = spot == main_spot
        function(current, *tile)

    _mesh_delegate(None, True, delegate_function, tiles, point)


@_call_direct
def mesh_focal(function):
    
    """
    Call a function on each tile of the mesh, with a :class:`bool` denoting whether it's the pointed tile prepended in the arguments.

    :param function:
        The function called.
    """

    return functools.partial(_mesh_focal, function)


def _mesh_light(focus_color, evade_color, 
                tiles, point):
    
    colors = (evade_color, focus_color)
    
    def function(current, *tile):
        color = colors[current]
        _text_paint(color, *tile)
    
    _mesh_focal(function, tiles, point)


@_call_direct
def mesh_light(focus_color: typing.Union[str, None] = None,
               evade_color: typing.Union[str, None] = None):
    
    """
    Paint the mesh's tiles depending on whether they are pointed.

    :param focus_color:
        The color used for the pointed tile.
    :param evade_color:
        The color used for the other tiles.
    """

    return functools.partial(_mesh_light, focus_color, evade_color)


def _mesh_point(focus_rune, evade_rune, tiles, point):

    runes = (evade_rune, focus_rune)

    def function(current, lines, point):
        rune = runes[current]
        size = len(rune)
        fill = size * ' '
        for (index, line) in enumerate(lines):
            text = fill if index else rune
            line[0:0] = text
        point[1] += size
    
    _mesh_focal(function, tiles, point)


@_call_direct
def mesh_point(focus_rune: str,
               evade_rune: str):
    
    """
    Prepend a rune to the pointed tile's first line and another to all others.

    All other lines have empty characters prepended to align with the first.

    :param focus_rune:
        The rune for the pointed tile.
    :param evade_rune:
        The rune for the other tiles.
    """

    return functools.partial(_mesh_point, focus_rune, evade_rune)


def _mesh_max(axis, size, 
              tiles, point):

    size -= 1

    cur_a = point[axis]

    all_a = set(map(operator.itemgetter(axis), tiles))
    min_a = builtins.min(all_a)
    max_a = builtins.max(all_a)

    cut_a = size // 2

    min_a = builtins.min(max_a - size, builtins.max(min_a, cur_a - cut_a))
    max_a = builtins.min(max_a, min_a + size)

    for spot in tuple(tiles):
        cur_a = spot[axis]
        if cur_a >= min_a and cur_a <= max_a:
            continue  
        del tiles[spot]


@_call_direct
def mesh_max(axis: int, 
             size: int):
    
    """
    Ensure there is most of a certain amount of tiles along an certain axis.

    :param axis:
        The axis considered for the operation.
    :param size:
        The maximum amount of tiles.
    """
    
    return functools.partial(_mesh_max, axis, size)


def _mesh_min(axis, just, size, fill,
              tiles, point):

    axis_oth = (axis - 1) % len(point) 

    all_a = set(map(operator.itemgetter(axis), tiles))
    min_a = builtins.min(all_a)
    max_a = builtins.max(all_a) 

    padding_size = size - (max_a - min_a)

    if not padding_size > 0:
        return

    if just == JustType.start:
        padding_size_start = 0
        padding_size_end = padding_size
    elif just == JustType.end:
        padding_size_start = padding_size
        padding_size_end = 0
    elif just == JustType.center:
        padding_size_start = padding_size // 2
        padding_size_end = padding_size - padding_size_start

    new_min_a = min_a - padding_size_start
    new_max_a = max_a + padding_size_end

    all_o = set(map(operator.itemgetter(axis_oth), tiles))
    min_o = builtins.min(all_o)
    max_o = builtins.max(all_o)

    new_range_a_start = range(new_min_a, min_a)
    new_range_a_end = range(max_a + 1, new_max_a + 1)

    for new_a in itertools.chain(new_range_a_start, new_range_a_end):
        for cur_o in range(min_o, max_o + 1):
            spot = [cur_o, cur_o]
            spot[axis] = new_a
            spot = tuple(spot)
            if spot in tiles:
                continue
            lines = fill(spot)
            tiles[spot] = (lines, [0, 0])


@_call_direct
def mesh_min(axis: int, 
             size: int, 
             just: JustType = JustType.start, 
             fill: typing.Callable[[typing.Tuple[int, int]], typing.List[typing.List[str]]] = lambda spot: [[]]):
    
    """
    Ensure there is most of a certain amount of tiles along an certain axis, aligned accordingly by ``fill``\ing with new tiles.

    :param axis:
        The axis considered for the operation.
    :param size:
        The minimum amount of tiles.
    :param just:
        Denotes justification to align with.
    :param fill:
        Used as :code:`(spot) -> lines` to fill empty space.
    """
    
    return functools.partial(_mesh_min, axis, just, size, fill)


def _mesh_grid_fill(tile_just_vertical, tile_rune_vertical,
                    tile_just_horizontal, tile_rune_horizontal,
                    tile_min_vertical, tile_min_horizontal,
                    push_top, push_bottom, push_left, push_right,
                    tiles, point):
    
    col_x_group = collections.defaultdict(list)
    col_y_group = collections.defaultdict(list)

    sizes_vertical = collections.defaultdict(int)
    sizes_horizontal = collections.defaultdict(int)

    for ((cur_y, cur_x), (lines, point)) in tiles.items():
        col_x_group[cur_y].append(cur_x)
        col_y_group[cur_x].append(cur_y)
        size_vertical = len(lines)
        sizes_vertical[cur_y] = builtins.max(sizes_vertical[cur_y], size_vertical, tile_min_vertical)
        size_horizontal = builtins.max(map(len, lines))
        sizes_horizontal[cur_x] = builtins.max(sizes_horizontal[cur_x], size_horizontal, tile_min_horizontal)

    bounds_x = {cur_y: (builtins.min(col_x), builtins.max(col_x)) for (cur_y, col_x) in col_x_group.items()}
    bounds_y = {cur_x: (builtins.min(col_y), builtins.max(col_y)) for (cur_x, col_y) in col_y_group.items()}

    spots = set()
    for (cur_y, (min_x, max_x)) in bounds_x.items():
        for may_x in range(min_x, max_x + 1):
            spot = (cur_y, may_x)
            spots.add(spot)
    for (cur_x, (min_y, max_y)) in bounds_y.items():
        for may_y in range(max_y, min_y, - 1):
            spot = (may_y, cur_x)
            spots.add(spot)

    min_y = builtins.min(bounds_x)
    max_y = builtins.max(bounds_x)
    all_y = range(max_y + push_top, min_y - 1 - push_bottom, - 1)

    min_x = builtins.min(bounds_y)
    max_x = builtins.max(bounds_y)
    all_x = range(min_x - push_left, max_x + 1 + push_right, 1)

    for spot in itertools.product(all_y, all_x):
        try:
            tile = tiles[spot]
        except KeyError:
            tile = tiles[spot] = ([[]], [0, 0])
        size_horizontal = sizes_horizontal[spot[1]]
        _text_min_horizontal(tile_just_horizontal, size_horizontal, tile_rune_horizontal, *tile)
        size_vertical = sizes_vertical[spot[0]]
        _text_min_vertical(tile_just_vertical, size_vertical, tile_rune_vertical, *tile)


@_call_direct
def mesh_grid_fill(tile_just_vertical  : JustType = JustType.start, 
                   tile_rune_vertical  : str      = ' ',
                   tile_just_horizontal: JustType = JustType.start, 
                   tile_rune_horizontal: str      = ' ',
                   tile_min_vertical   : int      = 0, 
                   tile_min_horizontal : int      = 0,
                   push_top            : int      = 0, 
                   push_bottom         : int      = 0, 
                   push_left           : int      = 0, 
                   push_right          : int      = 0):

    """
    Fill missing spots within a square so that all rows and columns align.

    For that to happen, tiles may need to grow to adjust for the size of others in each axis.

    :param tile_just_vertical:
        Denotes vertical justfication for growing each tile.
    :param tile_rune_vertical:
        The rune used when vertical growing each tile.
    :param tile_just_horizontal:
        Denotes horizontal justfication for growing each tile.
    :param tile_rune_horizontal:
        The rune used when horizontal growing each tile.
    :param tile_min_vertical:
        The minimum vertical size of each tile.
    :param tile_min_horizontal:
        The minimum horizontal size of each tile.
    :param push_top:
        The additional amount of horizontal groups of tiles injected over the final square.
    :param push_bottom:
        The additional amount of horizontal groups of tiles injected under the final square.
    :param push_left:
        The additional amount of vertical groups of tiles injected left of the final square.
    :param push_right:
        The additional amount of vertical groups of tiles injected right of the final square.
    """
    
    return functools.partial(
        _mesh_grid_fill, 
        tile_just_vertical, tile_rune_vertical,
        tile_just_horizontal, tile_rune_horizontal,
        tile_min_vertical, tile_min_horizontal,
        push_top, push_bottom, push_left, push_right
    )


def _mesh_grid_get_text_block_runes(get, spots, spot):

    (cur_y, cur_x) = spot

    got_current      = (cur_y    , cur_x    ) in spots
    got_top          = (cur_y + 1, cur_x    ) in spots
    got_left         = (cur_y    , cur_x - 1) in spots
    got_top_left     = (cur_y + 1, cur_x - 1) in spots

    # rune_top
    if got_current or got_top:
        yield get('horizontal')
    else:
        yield ' '

    yield None

    # rune_left
    if got_current or got_left:
        yield get('vertical')
    else:
        yield ' '
    
    yield None

    # rune_top_left
    if got_top and got_left:
        yield get('cross')
    elif got_current:
        if got_top_left:
            yield get('cross')
        elif got_top:
            yield get('cross_left')
        elif got_left:
            yield get('cross_top')
        else:
            yield get('top_left')
    elif got_top_left:
        if got_top:
            yield get('cross_bottom')
        elif got_left:
            yield get('cross_right')
        else:
            yield get('bottom_right')
    elif got_top:
        yield get('bottom_left')
    elif got_left:
        yield get('top_right')
    else:
        yield ' '


def _mesh_grid(runes, runes_color, tiles, point, **mesh_grid_fill_kwargs):

    @functools.lru_cache()
    def get_rune(name):
        rune = runes[name]
        if not runes_color is None:
            rune =_helpers.paint_rune(runes_color, rune)
        return rune

    spots = set(tiles)
    
    mesh_grid_fill.call(
        tiles, point, 
        **mesh_grid_fill_kwargs, 
        push_top = 0, push_bottom = 1, push_left = 0, push_right = 1
    )

    all_y = set(map(operator.itemgetter(0), tiles))
    min_y = builtins.min(all_y)

    all_x = set(map(operator.itemgetter(1), tiles))
    max_x = builtins.max(all_x)
    
    for (spot, tile) in tiles.items():
        text_block_runes = tuple(_mesh_grid_get_text_block_runes(get_rune, spots, spot))
        if spot[0] > min_y:
            _text_bloat_vertical(JustType.end, 1, ' ', *tile)
        _text_bloat_horizontal(JustType.end, 1, ' ', *tile)
        _text_block(*text_block_runes, None, None, None, *tile)

    for (spot, tile) in tuple(tiles.items()):
        (cur_y, cur_x) = spot
        if cur_y == min_y and cur_x == max_x:
            old_spot = (cur_y + 1, cur_x - 1)
            old_tile = tiles[old_spot]
            new_rune = tile[0][0][0]
            old_tile[0][- 1].append(new_rune)
        elif cur_y == min_y:
            old_spot = (cur_y + 1, cur_x)
            old_tile = tiles[old_spot]
            new_line = tile[0][0]
            old_tile[0].append(new_line)
        elif cur_x == max_x:
            old_spot = (cur_y, cur_x - 1)
            old_tile = tiles[old_spot]
            for (old_line, new_line) in zip(old_tile[0], tile[0]):
                old_line.append(new_line[0])
        else:
            continue
        del tiles[spot]
        

@_call_direct
def mesh_grid(runes: typing.Dict[str, str] = {
                'horizontal': '─',
                'vertical': '│',
                'cross': '┼',
                'cross_top': '┬',
                'cross_bottom': '┴',
                'cross_left': '├',
                'cross_right': '┤',
                'top_left': '┌',
                'top_right': '┐',
                'bottom_left': '└',
                'bottom_right': '┘'
              }, 
              runes_color: str = _colors.basic('black.lite'),
              **fill_kwargs):
    
    """
    Turns a mesh into a grid using.
     
    :param runes:
        The collection of runes used for drawing around each tile.
    :param runes_color:
        Used to paint all provided grid runes.

    Additional arguments get passed to :func:`.mesh_grid_fill` before the operation.
    """

    return functools.partial(_mesh_grid, runes, runes_color, **fill_kwargs)


def _mesh_head(axis, just, skip, min, get,
               tiles, point):

    dimensions = len(point)

    axis_oth = (axis - 1) % dimensions

    all_a = set(map(operator.itemgetter(axis), tiles))
    min_a = builtins.min(all_a)
    max_a = builtins.max(all_a)

    all_o = map(operator.itemgetter(axis_oth), tiles)

    size = builtins.max(min, max_a - min_a + 1)

    if axis:
        max_a = min_a + size
        range_a = range(min_a, max_a, + 1)
        if just is None:
            just = JustType.start
    else:
        min_a = max_a - size
        range_a = range(max_a, min_a, - 1)
        if just is None:
            just = JustType.end

    if just is JustType.start:
        max_o = builtins.max(all_o)
        cur_o = max_o + 1
    elif just is JustType.end:
        min_o = builtins.min(all_o)
        cur_o = min_o - 1
    else:
        raise ValueError(f'invalid just: {just}')

    for (cur_i, cur_a) in enumerate(range_a):
        if cur_i < skip:
            continue
        lines = get(cur_a)
        if lines is None:
            continue
        point = [0, 0]
        tile = (lines, point)
        spot = [cur_o] * dimensions
        spot[axis] = cur_a
        spot = tuple(spot)
        tiles[spot] = tile


@_call_direct
def mesh_head(axis: int, 
              get : typing.Callable[[int], typing.List[typing.List[str]]], 
              just: typing.Union[JustType, None] = None,
              skip: int = 0, 
              min : int = 0):
    
    """
    Add headers to the mesh.
    
    :param axis:
        The axis along which to add headers.
    :param get:
        Used as :code:`(index) -> lines` on each row or column for fetching its header.
    :param just:
        Denotes justification to align with. Default is left/top.
    :param skip:
        The amount of rows or columns to ignore at the beginning of the operation.
    :param min:
        The minimum amount of headers to be added. Can be used to avoid possible gaps.

    .. code-block::

        mesh_head(0, lambda index: [list(str(index))], min = 10)
        # skip is used to prevent adding a header for the headers row.
        mesh_head(1, lambda index: [list(str(index))], min = 10, skip = 1)
    """
    
    return functools.partial(_mesh_head, axis, just, skip, min, get)


def _mesh_flip(axis,
               tiles, point):
    
    flip = lambda iterable: tuple(value * - 1 for value in iterable)
    
    temp = {}
    for spot in tuple(tiles):
        tile = tiles.pop(spot)
        spot = flip(spot)
        temp[spot] = tile
    
    tiles.update(temp)

    point[:] = flip(point)


@_call_direct
def mesh_flip(axis: int):
    
    """
    Flip a mesh's axis.
    
    :param axis:
        The axis to flip.
    """
    
    return functools.partial(_mesh_flip, axis)



"""
General utilities.
"""

from . import _helpers


__all__ = ('paint',)


def paint(color: str, 
          value: str,
          *,
          force: bool = False) -> str:

    """
    Get the colored version of the value.
    """

    if force:
        value = _helpers.SGR.clean(value)

    return _helpers.paint_text(color, value)


def chain_funnels(*funnels):

    return _helpers.chain_functions(*funnels)


def split_lines(value):

    return _helpers.split_lines(value)


def join_lines(value):

    return _helpers.join_lines(value)


import typing
import enum
import threading

from . import _helpers
from . import _ansi
from . import _intel


__all__ = ('ClearMode', 'EraseMode', 'Cursor')


class ClearMode(enum.IntEnum):

    """
    ClearMode()
    
    Denotes how the cursor clears the screen.
    """

    #: Clear all runes right-and-down of the cursor.
    right  = 0
    #: Clear all runes left-and-up of the cursor.
    left   = 1
    #: Clear all runes.
    full   = 2
    #: Clear all runes and delete scrollback lines.
    extend = 3


class EraseMode(enum.IntEnum):

    """
    EraseMode()
    
    Denotes how the cursor erases the line.
    """

    #: Erase all visible runes right of the cursor.
    right  = 0
    #: Erase all visible runes left of the cursor.
    left   = 1
    #: Erase all visible runes.
    full   = 2


class Cursor:

    """
    Emulates the on-screen cursor.

    .. note::

        Coordinates here are ``1``-offset indexes, meaning they start from 1. 
        The position of the cursor on the top-left-most of the screen is :code:`(1, 1)`.
        
    :parma intel:
        Used for fetching parsed items from input and sending to output.

    .. code-block:: python
    
        io = IO(sys.stdin, sys.stdout)
        intel = Intel(io)
        cursor = Cursor(intel)
        location = cursor.locate()
    """

    __slots__ = ('_intel', '_hidden', '_lock')

    def __init__(self, 
                 intel: _intel.Intel):

        self._intel = intel

        self._hidden = _helpers.Atomic(self.hide, self.show)

        self._lock = threading.RLock()

    @property
    def hidden(self):

        """
        A context manager for hiding the cursor.

        .. code-block:: python

            with cursor.hidden:
                ...
        """

        return self._hidden

    def _send(self, *args, escape = False, **kwargs):

        function = _ansi.get_escape if escape else _ansi.get_control

        value = function(*args, **kwargs)

        self._intel.send(value)

    def _up(self, size):

        self._send('A', size)

    def up(self, size: typing.Union[int, None] = None):

        """
        Move up. No effect if at the edge of the screen.

        :param size:
            The amount of lines to move by. System default  is usually :code:`1`.
        """

        self._up(size)

    def _down(self, size):

        self._send('B', size)

    def down(self, size: typing.Union[int, None] = None):

        """
        Move down. No effect if at the edge of the screen.

        :param size:
            The amount of lines to move by. System default  is usually :code:`1`.
        """

        self._down(size)

    def _left(self, size):

        self._send('D', size)

    def left(self, size: typing.Union[int, None] = None):

        """
        Move left. No effect if at the edge of the screen.

        :param size:
            The amount of lines to move by. System default  is usually :code:`1`.
        """

        self._left(size)

    def _right(self, size):

        self._send('C', size)

    def right(self, size: typing.Union[int, None] = None):

        """
        Move right. No effect if at the edge of the screen.

        :param size:
            The amount of lines to move by. System default  is usually :code:`1`.
        """

        self._right(size)

    def _goto(self, x):

        self._send('G', x)

    def goto(self, x: int = None):

        """
        Move absolutely on line.

        :param x:
            The absolute coordinate to move to. System default  is usually :code:`1`.
        """

        self._goto(x)

    def _last(self, size):

        self._send('F', size)

    def last(self, size: typing.Union[int, None] = None):

        """
        Move up at line start.

        :param size:
            The amount of lines to move by. System default  is usually :code:`1`.
        """

        self._last(size)

    def _next(self, size):

        self._send('E', size)

    def next(self, size: typing.Union[int, None] = None):

        """
        Move down at line start.

        :param size:
            The amount of lines to move by. System default  is usually :code:`1`.
        """

        self._next(size)

    def _move(self, y, x):

        self._send('f', y, x)

    def move(self, y: typing.Union[int, None] = None, x: typing.Union[int, None] = None):

        """
        Move absolutely on screen.

        :param y:
            The absolute vertical coordinate to move to. System default  is usually :code:`1`.
        :param x:
            The absolute horizontal coordiante to move to. System default  is usually :code:`1`.
        """
        
        self._move(y, x)

    def _clear(self, mode):

        self._send('J', mode)

    def clear(self, mode: ClearMode = None):

        """
        Clear the screen.

        :param mode:
            Denotes which part of the screen to clear. System default  is usually :attr:`ClearMode.right`.
        """

        self._clear(mode)

    def _erase(self, mode):

        self._send('K', mode)

    def erase(self, mode: EraseMode = None):

        """
        Erase the line.

        :param mode:
            Denotes which part of the screen to clear. System default  is usually :attr:`EraseMode.right`.
        """

        self._erase(mode)

    def _save(self):

        self._send('7', escape = True)

    def save(self):

        """
        Save the current position. Consequent uses will overwrite previous ones.
        """

        self._save()

    def _load(self):

        self._send('8', escape = True)

    def load(self):

        """
        Move to the last-saved position. Consequent uses will have the same result.
        """

        self._load()

    def _hide(self):

        self._send('l', '25', private = True)

    def hide(self):

        """
        Become invisible.
        """

        self._hide()

    def _show(self):

        self._send('h', '25', private = True)

    def show(self):

        """
        Become visible.
        """

        self._show()

    @_helpers.ctxmethod(lambda self: self._lock)
    def _locate(self):

        self._send('n', '6')

        def check(code):
            return isinstance(code, _ansi.Sequence) and code.rune == 'R'
        
        code = self._intel.read(check = check)
    
        point = tuple(map(int, code.args))

        return point

    def locate(self) -> typing.Tuple[int, int]:

        """
        Get the current absolute vertical and horizontal coordinates.

        :returns:
            ``(vertical, horizontal)``
        """

        point = self._locate()

        return point

    _max_y = _max_x = 9999

    @_helpers.ctxmethod(lambda self: self._lock)
    def _measure(self):

        self._save()

        self._move(self._max_y, self._max_x)

        point = self._locate()

        self._load()

        return point

    def measure(self) -> typing.Tuple[int, int]:

        """
        Get the maximum possible vertical and horizontal coordinates.

        :returns:
            ``(vertical, horizontal)``
        """

        with self._hidden:
            point = self._measure()

        return point


import typing
import threading

from . import _helpers
from . import _cursor
from . import _render


__all__ = ('Screen',)


_type_Screen_draw_sketch = typing.Callable[[], typing.Tuple[_render._type_Render_draw_lines, _render._type_Render_draw_point]]


class Screen:

    """
    Encapsulates the visualization logic.

    :param render:
        Used for drawing and cursor placement.
    :param cursor:
        Used for hidding the cursor between draws.
    """

    __slots__ = ('_cursor', '_render', '_lock')

    def __init__(self, 
                 render: _render.Render, 
                 cursor: _cursor.Cursor):

        self._cursor = cursor
        self._render = render

        self._lock = threading.Lock()

    @_helpers.ctxmethod(lambda self: self._lock)
    @_helpers.ctxmethod(lambda self: self._cursor.hidden)
    def _print(self, sketch, re, **kwargs):

        if re:
            self._render.back()

        info = sketch()

        self._render.draw(*info, **kwargs)

    def print(self, 
              sketch: _type_Screen_draw_sketch,
              re: bool,
              *,
              clean: bool = True,
              learn: bool = True):
        
        """
        Fetch drawing information and use it to print.

        :param sketch:
            Used after potentially placing the cursor to its initial position. Should return ``(lines, point)`` used for drawing.
        :param re:
            Whether to return the cursor to its initial position before attempting to draw.
        :param clean:
            Same as :paramref:`.Render.draw.clean`.
        :param learn:
            Same as :paramref:`.Render.draw.learn`.

        The reason a callable is required instead of ``(lines, point)``, is to allow the original cursor information to be true on every call.
        """
        
        self._print(sketch, re, clean = clean, learn = learn)






"""
Combinations of :class:`~.mutates.Mutate`\ s, :class:`~.visuals.Visual`\ s and 
:class:`~.handle.Handle` to create interactive units that can be resolved into expected values. 
"""

import typing
import contextvars
import itertools
import functools
import math
import datetime

from . import _core
from . import _helpers
from . import _system
from . import _colors
from . import _mutates
from . import _visuals
from . import _controls
from . import _handle
from . import _stage
from . import _funnels
from . import _searches
from . import _theme


__all__ = ('WidgetError', 'Widget', 'start', 
           'BaseText', 'Input', 'Numeric', 'Conceal', 'AutoSubmit', 'Inquire', 
           'BaseMesh', 'BaseList', 'Select', 'Basket', 'Count', 'DateTime',
           'Form')


class Abort(_helpers.InfoErrorMixin, Exception):

    """
    Raised when an invokation needs to stop. May include a message.
    """

    __slots__ = ()



class Escape(Exception):

    """
    Raised when the users pressed the ``Esc`` button.
    """

    __slots__ = ()


_type_Widget_init_mutate    = _mutates.Mutate
_type_Widget_init_visual    = _visuals.Visual
_type_Widget_init_callback  = _handle._type_Handle_init_callback
_type_Widget_init_delegate  = typing.Callable[[_core.Event], bool]
_type_Widget_init_validate  = typing.Callable[[typing.Any], None]
_type_Widget_init_escapable = bool

_type_Widget_invoke_event = _core.Event
_type_Widget_invoke_info  = _core._type_ansi_parse_return


class Widget:

    """
    A team of mutate, handle and visual working together.

    :param mutate:
        The underlying mutate for changing data.
    :param visual:
        The underlying visual for fetching drawing information.
    :param callback:
        Used as :paramref:`.handle.Handle.callback`.
    :param delegate:
        Used with ``(event)`` and decides whether to continue the invokation.
    :param validate:
        Used with ``(result)`` upon submission, forbidden by raising :exc:`.Abort`.
    """

    _product_mark = object()

    __slots__ = ('_mutate', '_handle', '_visual', '_delegate', '_validate', '_escapable', '_product')

    def __init_subclass__(cls, controls = (), **kwargs):

        pre_controls = getattr(cls, '_controls', ())
        
        cls._controls = (*pre_controls, *controls)

        super().__init_subclass__(**kwargs)

    def __init__(self, 
                 mutate   : _type_Widget_init_mutate, 
                 visual   : _type_Widget_init_visual, 
                 callback : _type_Widget_init_callback  = None, 
                 delegate : _type_Widget_init_delegate  = None,
                 validate : _type_Widget_init_validate  = None,
                 escapable: _type_Widget_init_escapable = False):
        
        self._delegate = delegate
        self._validate = validate
        self._product = self._product_mark

        self._mutate = mutate

        handle = _handle.Handle(mutate, callback = callback)

        for control in self._controls:
            handle.add(control)

        self._handle = handle
        self._visual = visual

        self._escapable = escapable

    @property
    def mutate(self):

        """
        The underlying mutate.
        """

        return self._mutate
    
    def _prepare(self, value):

        return value
    
    def _resolve(self):

        raise NotImplemented()
    
    def resolve(self):

        """
        Get the resolved value.
        """

        value = self._product

        if value is self._product_mark:
            value = self._resolve()

        value = self._prepare(value)

        return value
    
    def _invoke_validate(self):

        validate = self._validate

        if validate is None:
            return

        value = self._product

        if value is self._product_mark:
            value = self._product = self._resolve()

        validate(value)
        
        self._product = self._product_mark

    def _invoke(self, event, *args, **kwargs):

        if self._delegate and not self._delegate(event):
            return

        try:
            self._handle.invoke(event, *args, **kwargs)
        except _core.Terminate:
            self._invoke_validate(); raise
    
    def invoke(self, 
               event: _type_Widget_invoke_event, 
               info : _type_Widget_invoke_info):
        
        """
        Invoke the underlying handle.

        :param event:
            Dictates the type of operation.
        :param info:
            Contains details used during the operation.

        .. note::

            If :paramref:`.delegate` is provided and returns :code:`False`, nothing happens.
        """

        if self._escapable and event == _core.Event.escape:
            raise Escape()

        self._invoke(event, info)
    
    def _sketch(self, *args, **kwargs):

        return self._visual.get(*args, **kwargs)
    
    def sketch(self, 
               *args, 
               **kwargs) -> _visuals._type_Visual_get_return:

        """
        Get the lines and point used for rendering the widget's state.

        Additional arguments are passed to the underlying :meth:`.Visual.get` method.

        :return:
            The ``(lines, point)`` that can be used to draw.
        """
        
        return self._sketch(*args, **kwargs)


def _start_variant_parse(parse, value):
    if parse:
        lines = _helpers.split_lines(value)
        point_y = len(lines) - 1
        point_x = len(lines[point_y])
        point = [point_y, point_x]
        value = (lines, point)
    return value


_type_start_get_actor_contextual_parse   = bool
_type_start_get_actor_contextual_context = contextvars.ContextVar


def _start_get_actor_contextual(parse  : _type_start_get_actor_contextual_parse,
                                context: _type_start_get_actor_contextual_context):

    def wrapper(*args, **kwargs):
        value = context.get()
        return _start_variant_parse(parse, value)
    
    return wrapper


_type_start_start_get_actor_static_parse = _type_start_get_actor_contextual_parse
_type_start_start_get_actor_static_value = typing.Union[str, typing.List[typing.List[str]]]


def _start_get_actor_static(parse: _type_start_start_get_actor_static_parse, 
                            value: _type_start_start_get_actor_static_value):

    context = contextvars.ContextVar('state')

    context.set(value)

    wrapper = _start_get_actor_contextual(parse, context)

    return wrapper


_type_start_start_get_actor_dynamic_parse = _type_start_get_actor_contextual_parse
_type_start_start_get_actor_dynamic_fetch = typing.Callable[[typing.Any], typing.Tuple[_visuals._type_Text_link_lines, _visuals._type_Text_link_point]]


def _start_get_actor_dynamic(parse: _type_start_start_get_actor_dynamic_parse, 
                             fetch: _type_start_start_get_actor_dynamic_fetch):

    context = contextvars.ContextVar('state')

    def updater(*args, **kwargs):
        context.set(fetch(*args, **kwargs))
    
    wrapper = _start_get_actor_contextual(parse, context)

    return wrapper, updater


_type_start_get_actor_dichotomic_parse = _type_start_get_actor_contextual_parse
_type_start_get_actor_dichotomic_value = typing.Union[_type_start_start_get_actor_static_value, _type_start_start_get_actor_dynamic_fetch]


def _start_get_actor_dichotomic(parse: _type_start_get_actor_dichotomic_parse, 
                                value: _type_start_get_actor_dichotomic_value):

    if callable(value):
        wrapper, updater = _start_get_actor_dynamic(parse, value)
    else:
        wrapper = _start_get_actor_static(parse, value)
        updater = _helpers.noop

    return wrapper, updater


_start_warn_reset_lines = [[]]

_type_start_multi_pre  = _stage._type_get_multi_maybe
_type_start_multi_aft  = bool
_type_start_widget     = Widget
_type_start_show       = typing.Union[None, str]
_type_start_mark       = typing.Union[None, str]
_type_start_mark_color = str
_type_start_info       = _type_start_get_actor_dichotomic_value
_type_start_info_parse = _type_start_get_actor_dichotomic_parse
_type_start_hint       = _type_start_get_actor_dichotomic_value
_type_start_hint_parse = _type_start_get_actor_dichotomic_parse
_type_start_site       = _stage._type_get_site
_type_start_reply      = typing.Union[None, typing.Callable[[Widget, typing.Any], str]]


def _start(multi_pre : _type_start_multi_pre, 
           multi_aft : _type_start_multi_aft,
           widget    : _type_start_widget, 
           show      : _type_start_show       = None,
           mark      : _type_start_mark       = '? ',
           mark_color: _type_start_mark_color = _colors.basic('yellow'),
           info      : _type_start_info       = '',
           info_parse: _type_start_info_parse = True,
           hint      : _type_start_hint       = '', 
           hint_parse: _type_start_hint_parse = True,
           site      : _type_start_site       = 'body',
           reply     : _type_start_reply      = None):
    
    """
    Start a widget and return its resolve result upon submission.

    :param multi_pre:
        Whether the body should start on a separate line during use.
    :param multi_aft:
        Whether the reply should start on a separate line.
    :param widget:
        The widget to use.
    :param show:
        Printed before anything else (uncontrollable).
    :param mark:
        Prepended to the ``show``.
    :parm mark_color:
        The color to paint ``mark`` with.
    :param info:
        The info for the stage. If callable, takes ``(widget, event_name, event_info)``.
    :param info_parse:
        Whether ``info`` is given (or, if callable returns) as a :class:`str` or ``(lines, point)``.
    :param hint:
        The hint for the stage. If callable, takes ``(widget, event_name, event_info)``.
    :param hint_parse:
        Whether ``hint`` is given (or, if callable returns) as a :class:`str` or ``(lines, point)``.
    :param site:
        Can be :code:`'info'` or :code:`'body'` to place the cursor on the info or widget body respectively.
    :param reply:
        Called upon successful submission with ``(widget, result)``. Should return a class:`str` that is used as a response.
    """

    if not show is None:
        if not mark is None:
            if mark_color:
                mark = _helpers.paint_text(mark_color, mark)
            show = mark + show
        show_get = _start_get_actor_static(True, show)
        _system.screen.print(show_get, False, learn = False)

    info_get, info_update = _start_get_actor_dichotomic(info_parse, info)
    hint_get, hint_update = _start_get_actor_dichotomic(hint_parse, hint)
    body_get = widget.sketch

    multi_pre_maybe = multi_pre
    multi_pre_force = bool(show)

    visual = _stage.get(multi_pre_maybe, multi_pre_force, site, info_get, hint_get, body_get)

    def sketch(*args, **kwargs):
        return visual.get(*args, **kwargs)
    
    update = _helpers.chain_functions(info_update, hint_update)

    memory = result = None
    
    def invoke(*args, **kwargs):
        nonlocal memory, result
        memory = widget.mutate.get_state()
        try:
            try:
                widget.invoke(*args, **kwargs)
            except _core.Terminate:
                result = widget.resolve()
                raise
            else:
                _stage.warn(_start_warn_reset_lines)
                update(widget, *args, **kwargs)
        except _mutates.Error:
            _system.io.ring()
            raise _core.SkipDraw()
        except Abort as error:
            widget.mutate.set_state(memory)
            message = error.text
            if not message is None:
                lines = _helpers.split_lines(message)
                _stage.warn(lines)
            _system.io.ring()

    update(widget, None, None) # emulate

    try:
        _system.console.start(sketch, invoke)
    except BaseException:
        _system.cursor.clear(); raise

    def sketch():
        if reply is None:
            lines = []
        else:
            value = reply(widget, result)
            lines = _start_variant_parse(True, value)[0]
        if multi_aft and not show is None:
            lines.insert(0, [])
        lines.append([])
        return (lines, None)
    _system.screen.print(sketch, True)

    return result


start = _start


_type_BaseText_init_lines        = _mutates._type_Text_init_lines 
_type_BaseText_init_point        = _mutates._type_Text_init_point
_type_BaseText_init_multi        = bool
_type_BaseText_init_callback     = _type_Widget_init_callback
_type_BaseText_init_delegate     = _type_Widget_init_delegate
_type_BaseText_init_validate     = _type_Widget_init_validate
_type_BaseText_init_escapable    = _type_Widget_init_escapable
_type_BaseText_init_funnel_enter = _visuals._type_Text_init_funnel_enter
_type_BaseText_init_funnel_leave = _visuals._type_Text_init_funnel_leave


_BaseText_controls = (
    _controls.text_insert,
    _controls.text_move_left,
    _controls.text_move_right,
    _controls.text_move_up,
    _controls.text_move_down,
    _controls.text_delete_left,
    _controls.text_delete_right,
    _controls.text_newline
)


class BaseText(Widget, controls = _BaseText_controls):

    """
    Base for text-like widgets.

    :param lines:
        Same as :paramref:`.mutates.Text.lines`.
    :param point:
        Same as :paramref:`.mutates.Text.point`.
    :param multi:
        Whether to support multiple lines. If so, leaving 2 empty lines submits.
    :param callback:
        Same as :paramref:`.Widget.callback`.
    :param delegate:
        Same as :paramref:`.Widget.delegate`.
    :param validate:
        Same as :paramref:`.Widget.validate`.
    :param funnel_enter:
        Same as :paramref:`.visuals.Text.funnel_enter`.
    :param funnel_leave:
        Same as :paramref:`.visuals.Text.funnel_leave`.
    """

    __slots__ = ()

    def __init__(self, 
                 lines       : _type_BaseText_init_lines, 
                 point       : _type_BaseText_init_point,
                 multi       : _type_BaseText_init_multi        = False,
                 callback    : _type_BaseText_init_callback     = None,
                 delegate    : _type_BaseText_init_delegate     = None,
                 validate    : _type_BaseText_init_validate     = None,
                 escapable   : _type_BaseText_init_escapable    = False,
                 funnel_enter: _type_BaseText_init_funnel_enter = None, 
                 funnel_leave: _type_BaseText_init_funnel_leave = None):

        mutate = _mutates.Text(lines, point)

        def visual_get(*args):
            return (mutate.lines, mutate.point)

        visual = _visuals.Text(visual_get, funnel_enter, funnel_leave)

        handle = _handle.Handle()
        
        if not multi:
            @handle.add
            @_controls.get((_handle.EventType.enter, _core.Event.enter))
            def handle_entry(info):
                raise _core.Terminate()
            
        callback = _helpers.chain_functions(callback, handle.invoke)

        super().__init__(
            mutate, 
            visual,
            callback = callback,
            delegate = delegate,
            validate = validate,
            escapable = escapable
        )


_type_Input_init_value = str
_type_Input_init_index = int


class Input(BaseText):

    """
    A text editor.

    It resolves into a :class:`str`.

    :param value:
        Converted into :paramref:`.BaseText.lines`.
    :param index:
        Where to place in the cursor in relation to its position along the continous text. Converted into :paramref:`.BaseText.point`. 

    Arguments directly passed to super-class:
    
        - :paramref:`~.BaseText.lines` - Created from :paramref:`.value`.
        - :paramref:`~.BaseText.point` - Created from :paramref:`.index`.

    |theme| :code:`'widgets.Input'`.
    """

    __slots__ = ()

    @_theme.add('widgets.Input')
    def __init__(self, 
                 *args,
                 value: _type_Input_init_value = _helpers.auto, 
                 index: _type_Input_init_index = _helpers.auto,
                 **kwargs):
        
        if value is _helpers.auto:
            value = ''
        
        if index is _helpers.auto:
            index = len(value)

        lines = _helpers.split_lines(value)

        (y, x) = _helpers.text_index_to_point(lines, index)

        point = [y, x]

        super().__init__(lines, point, *args, **kwargs)

    def _resolve(self):

        value = _helpers.join_lines(self._mutate.lines)

        return value
    

_type_Numeric_init_value                 = int
_type_Numeric_init_decimal               = bool
_type_Numeric_init_zfill                 = int
_type_Numeric_init_invalid_value_message = str


class Numeric(Input):

    """
    A text editor that only allows submission with numeric values.

    It resolves into a :class:`int` or :class:`float`.

    :param value:
        Converted into :paramref:`.Input.value`.
    :param decimal:
        Whether to allow (finite) decimal values.
    :param zfill:
        Amount of 0-fill to have.
    :param invalid_value_message:
        Used as template to raise :exc:`.Abort` when the value is invalid.

    Arguments directly passed to super-class:
    
        - :paramref:`~.Input.value` - Created from :paramref:`.value`.

    |theme| :code:`'widgets.Numeric'`.
    """

    @staticmethod
    def _transform_float(abort_message, value):
        try:
            value = float(value)
        except ValueError:
            raise Abort(abort_message)
        if math.isinf(value):
            raise Abort(abort_message)
        return value
    
    @staticmethod
    def _transform_int(abort_message, value):
        try:
            value = int(value)
        except ValueError:
            raise Abort(abort_message)
        return value
    
    __slots__ = ('_transform', '_transform_abort_message')

    @_theme.add('widgets.Numeric')
    def __init__(self, 
                 *args,
                 value                : _type_Numeric_init_value                 = _helpers.auto,
                 decimal              : _type_Numeric_init_decimal               = False,
                 zfill                : _type_Numeric_init_zfill                 = 0,
                 invalid_value_message: _type_Numeric_init_invalid_value_message = 'invalid {name}',
                 **kwargs):
        
        super_cls = self.__class__.__mro__[1]

        funnel_leave = _helpers.get_function_arg_safe(super_cls, 'funnel_leave', kwargs, pop = True)
        
        if not value is _helpers.auto:
            value = str(value)

        if decimal:
            transform = self._transform_float
            abort_message_space = {'name': 'float'} 
        else:
            transform = self._transform_int
            abort_message_space = {'name': 'int'}

        self._transform = transform
        self._transform_abort_message = None if invalid_value_message is None else invalid_value_message.format(**abort_message_space)

        funnel_leave_group = []
        funnel_leave_entry = _funnels.text_min_horizontal(_funnels.JustType.end, zfill, '0')
        funnel_leave_group.append(funnel_leave_entry)
        funnel_leave = _helpers.chain_functions(*funnel_leave_group, funnel_leave)

        super().__init__(
            *args,
            value = value,
            funnel_leave = funnel_leave,
            **kwargs
        )

    def _resolve(self):

        value = super()._resolve()

        if value == '-':
            value = '0'

        if value.endswith('.'):
            value = f'{value}0'

        value = self._transform(self._transform_abort_message, value)

        return value


_type_Conceal_init_rune  = str
_type_Conceal_init_color = typing.Union[None, int]


class Conceal(Input):

    """
    A text editor that replaces all characters with a rune.

    It resolves into a :class:`str`.

    :param rune:
        The rune to replace all others with.
    :param color:
        The color to paint the rune with.

    |theme| :code:`'widgets.Conceal'`.
    """

    __slots__ = ()

    @_theme.add('widgets.Conceal')
    def __init__(self, 
                 *args, 
                 rune : _type_Conceal_init_rune  = '*', 
                 color: _type_Conceal_init_color = None, 
                 **kwargs):

        super_cls = self.__class__.__mro__[1]

        funnel_leave = _helpers.get_function_arg_safe(super_cls, 'funnel_leave', kwargs, pop = True)

        funnel_leave_group = []
        funnel_leave_entry = _funnels.text_replace(rune)
        funnel_leave_group.append(funnel_leave_entry)
        if not color is None:
            funnel_leave_entry = _funnels.text_paint(color)
            funnel_leave_group.append(funnel_leave_entry)
        funnel_leave = _helpers.chain_functions(*funnel_leave_group, funnel_leave)

        super().__init__(*args, funnel_leave = funnel_leave, **kwargs)


_type_AutoSubmit_init_evaluate  = typing.Callable[[str], None]
_type_AutoSubmit_init_validate  = typing.Callable[[str], bool]
_type_AutoSubmit_init_default   = typing.Any
_type_AutoSubmit_init_transform = typing.Union[None, typing.Callable[[str], str]]


class AutoSubmit(Input):

    """
    A text editor that automatically submits a upon valid insertion.

    Resolves into a :class:`str`.

    :param evaluate:
        Used with ``(result)`` before insertion, which can be prevented by raising :exc:`.Abort`.
    :param validate:
        Used with ``(result)`` after to decide whether to submit.
    :param default:
        Used instead when attempting to submit without a value.
    :param transform:
        Used with ``(result)`` and should return a new ``result`` used for checks.

    Arguments directly passed to super-class:
    
        - :paramref:`~.Input.validate` - Set to :code:`None`.
    """

    __slots__ = ()

    _default_mark = _helpers.auto

    def __init__(self, 
                 evaluate : _type_AutoSubmit_init_evaluate,
                 validate : _type_AutoSubmit_init_validate,
                 *args,
                 default  : _type_AutoSubmit_init_default   = _default_mark, 
                 transform: _type_AutoSubmit_init_transform = None,
                 **kwargs):
        
        super_cls = self.__class__.__mro__[1]

        callback = _helpers.get_function_arg_safe(super_cls, 'callback', kwargs, pop = True)
    
        handle = _handle.Handle()

        @handle.add
        @_controls.get((_handle.EventType.enter, _core.Event.enter))
        def _control_submit_enter(info):
            if _helpers.check_lines(self._mutate.lines) or default is self._default_mark:
                raise Abort(None)
            self._product = default
            raise _core.Terminate()

        _state = NotImplemented
        
        @handle.add
        @_controls.get((_handle.EventType.enter, _core.Event.insert))
        def _control_insert_enter(info):
            nonlocal _state
            _state = self._mutate.get_state()

        @handle.add
        @_controls.get((_handle.EventType.leave, _core.Event.insert))
        def _control_insert_leave(info):
            value = super(AutoSubmit, self)._resolve()
            if not transform is None:
                value = transform(value)
            try:
                evaluate(value)
            except Abort:
                self._mutate.set_state(_state); raise
            if not validate(value):
                return
            self._product = value
            raise _core.Terminate()
        
        callback = _helpers.chain_functions(callback, handle.invoke)

        super().__init__(
            *args, 
            validate = None,
            callback = callback,
            **kwargs
        )
            

_type_Inquire_init_options  = typing.Dict[str, typing.Any]
_type_Inquire_init_tranform = _type_AutoSubmit_init_transform
    

class Inquire(AutoSubmit):

    """
    A text editor that submits upon insertion of an option.

    Resolves into the value of the matching option.

    :param options:
        The possible options. Attempting to insert a rune that does not lead to an option is forbidden. 
    :param transform:
        Used on all options and the current value when comparing.

    Arguments directly passed to super-class:
    
        - :paramref:`~.AutoSubmit.evaluate` - Created using :paramref:`.options`.
        - :paramref:`~.AutoSubmit.validate` - Created using :paramref:`.options`.

    |theme| :code:`'widgets.Inquire'`.
    """

    __slots__ = ('_options',)

    @_theme.add('widgets.Inquire')
    def __init__(self, 
                 *args,
                 options  : _type_Inquire_init_options  = {'y': True, 'n': False}, 
                 transform: _type_Inquire_init_tranform = str.lower,
                 **kwargs):

        if not transform is None:
            options = dict(zip(map(transform, options), options.values()))

        self._options = options
        
        def evaluate(value):
            for option in options:
                if option.startswith(value):
                    return
            raise Abort(None)
        
        def validate(value):
            return value in options
        
        super().__init__(
            evaluate, 
            validate, 
            *args, 
            transform = transform,
            **kwargs
        )

    def _prepare(self, value):

        value = super()._prepare(value)

        value = self._options[value]

        return value
    
    
def _focus_nil(spot):

    return False


def _focus_all(spot): 

    return True


def _get_mesh_point(axis, index, default = 0):

    return _helpers.get_axis_point(2, axis, default, index)


def _get_mesh_spot(*args, **kwargs):

    point = _get_mesh_point(*args, **kwargs)

    return tuple(point)


_type_BaseMesh_init_search       = _mutates._type_Mesh_init_score
_type_BaseMesh_init_create       = _mutates._type_Mesh_init_create
_type_BaseMesh_init_tiles        = _mutates._type_Mesh_init_tiles
_type_BaseMesh_init_point        = _mutates._type_Mesh_init_point
_type_BaseMesh_init_clean        = _mutates._type_Mesh_init_clean
_type_BaseMesh_init_scout        = _mutates._type_Mesh_init_scout
_type_BaseMesh_init_rigid        = _mutates._type_Mesh_init_rigid
_type_BaseMesh_init_focus        = typing.Union[bool, typing.Callable[[_core.Event], bool] ]
_type_BaseMesh_init_callback     = _type_Widget_init_callback
_type_BaseMesh_init_delegate     = _type_Widget_init_delegate
_type_BaseMesh_init_validate     = _type_Widget_init_validate
_type_BaseMesh_init_escapable    = _type_Widget_init_escapable
_type_BaseMesh_init_funnel_enter = _visuals._type_Mesh_init_funnel_enter
_type_BaseMesh_init_funnel_leave = _visuals._type_Mesh_init_funnel_leave


_BaseMesh_controls = (
    _controls.mesh_move_left,
    _controls.mesh_move_right,
    _controls.mesh_move_up,
    _controls.mesh_move_down,
    _controls.mesh_insert,
    _controls.mesh_delete,
    _controls.mesh_enter
)


class BaseMesh(Widget, controls = _BaseMesh_controls):

    """
    A mesh traverser.

    :param search:
        Same as :paramref:`.mutates.Mesh.score`.
    :param create:
        Same as :paramref:`.mutates.Mesh.create`.
    :param tiles:
        Same as :paramref:`.mutates.Mesh.tiles`. Values should be :class:`Widget` instances.
    :param point:
        Same as :paramref:`.mutates.Mesh.point`.
    :param clean:
        Same as :paramref:`.mutates.Mesh.clean`.
    :param scout:
        Same as :paramref:`.mutates.Mesh.scout`.
    :param rigid:
        Same as :paramref:`.mutates.Mesh.rigid`.
    :param focus:
        When :class:`bool`, it is used as the initial focus state. The state can be set to :code:`True` with :attr:`.Event.indent` while unfocused, and to :code:`False` by submitting the focused tile.
        When :class:`~typing.Callable`, it is used on each invokation to determine whether to delegate it to the currently pointed-at tile instead.
    :param callback:
        Same as :paramref:`.Widget.callback`.
    :param delegate:
        Same as :paramref:`.Widget.delegate`.
    :param validate:
        Same as :paramref:`.Widget.validate`.
    :param escapable:
        Same as :paramref:`.Widget.escapable`.
    :param funnel_enter:
        Same as :paramref:`.visuals.Mesh.funnel_enter`.
    :param funnel_leave:
        Same as :paramref:`.visuals.Mesh.funnel_leave`.
    """

    __slots__ = ('_focus',)

    def __init__(self, 
                 tiles       : _type_BaseMesh_init_tiles        = _helpers.auto, 
                 point       : _type_BaseMesh_init_point        = _helpers.auto, 
                 create      : _type_BaseMesh_init_create       = None,
                 search      : _type_BaseMesh_init_search       = _searches.fuzzy,
                 clean       : _type_BaseMesh_init_clean        = False,
                 scout       : _type_BaseMesh_init_scout        = None, 
                 rigid       : _type_BaseMesh_init_rigid        = False,
                 focus       : _type_BaseMesh_init_focus        = False,
                 callback    : _type_BaseMesh_init_callback     = None,
                 delegate    : _type_BaseMesh_init_delegate     = None,
                 validate    : _type_BaseMesh_init_validate     = None,
                 escapable   : _type_BaseMesh_init_escapable    = True,
                 funnel_enter: _type_BaseMesh_init_funnel_enter = None, 
                 funnel_leave: _type_BaseMesh_init_funnel_leave = None):
        
        if tiles is _helpers.auto:
            tiles = ()

        tiles = dict(tiles)
        
        if point is _helpers.auto:
            try:
                point_spot = min(tiles)
            except ValueError:
                point = [0, 0]
            else:
                point = list(point_spot)

        mutate = _mutates.Mesh(search, scout, rigid, create, clean, tiles, point)

        def visual_get(*args):
            tiles = {}
            for (vis_spot, cur_spot) in mutate.vision.items():
                try:
                    tile = mutate.tiles[cur_spot]
                except KeyError:
                    continue
                tiles[vis_spot] = tile.sketch(*args)
            point = mutate.point
            return (tiles, point)

        visual = _visuals.Mesh(visual_get, funnel_enter, funnel_leave)

        self._focus = focus

        super().__init__(
            mutate, 
            visual,
            escapable = escapable,
            validate = validate,
            delegate = delegate,
            callback = callback
        )

    @property
    def focus(self):

        return self._focus

    @property
    def _focusable(self):

        return not callable(self._focus)
    
    def _invoke_blear(self, *args, **kwargs):

        super()._invoke(*args, **kwargs)

    def _handle_focus(self):

        self._focus = not self._focus
        
    def _invoke_focus(self, *args, **kwargs):

        tile = self._mutate.cur_tile

        try:
            tile.invoke(*args, **kwargs)
        except _core.Terminate:
            if self._focusable:
                self._handle_focus()
                if not self._focus:
                    return
            raise

    def _invoke(self, event, *args, **kwargs):

        switcher = self._focusable
        delegate = self._focus if switcher else self._focus(event)

        if delegate:
            self._invoke_focus(event, *args, **kwargs)
        elif switcher and event is _core.Event.indent:
            self._handle_focus()
        else:
            self._invoke_blear(event, *args, **kwargs)


_type_BaseList_init_tiles       = typing.Union[typing.List[Widget], _type_BaseMesh_init_tiles]
_type_BaseList_init_axis        = int
_type_BaseList_init_index       = int
_type_BaseList_init_label       = typing.Union[None, typing.Callable[[int], str]]
_type_BaseList_init_view_max    = typing.Union[None, int]
_type_BaseList_init_focus_color = typing.Union[None, str]
_type_BaseList_init_focus_mark  = typing.Union[None, str]
_type_BaseList_init_evade_color = typing.Union[None, str]
_type_BaseList_init_evade_mark  = typing.Union[None, str]
_type_BaseList_init_fill        = bool


_BaseList_controls = (
    _controls.mesh_move_up_reverse,
    _controls.mesh_move_down_reverse,
)


class BaseList(BaseMesh, controls = _BaseList_controls):

    """
    Base for list-like widgets.

    Resolves to a :class:`int` (the pointed index).

    :param axis:
        Denotes the axis across which movement is allowed.
    :param tiles:
        Same as :paramref:`.BaseMesh.tiles`, but it can be any iterable.
    :param index:
        Used as the :paramref:`.axis`\th element of :paramref:`.BaseMesh.point`.
    :param label:   
        Used as ``(index, tile)`` and returns a label to place on the right. 
    :param view_max:
        Maximum amount of elemnts visible at once.
    :param focus_color:
        The color to paint the current tile with.
    :param focus_mark:  
        The rune to prepend to the current tile.
    :param evade_color:
        The color to paint the non-current tiles with.
    :param evade_mark:
        The rune to prepend to the non-current tiles.
    :param fill:
        Whether to fill empty room.

    Arguments directly passed to super-class:
    
        - :paramref:`~.BaseMesh.tiles` - Created using :paramref:`.tiles`.
        - :paramref:`~.BaseMesh.point` - Created using :paramref:`.index`.

    .. warning::
        For the sake of matching each spot's index to it's ``axis``\ -th value, vertical movement and vision are
        adjusted so that indexes increment downward instead of upward (which is how :class:`.BaseMesh` behaves normally).
    """
        
    __slots__ = ('_axis',)

    def __init__(self,
                 *args,
                 axis       : _type_BaseList_init_axis        = 0,
                 tiles      : _type_BaseList_init_tiles       = _helpers.auto,
                 index      : _type_BaseList_init_index       = 0,
                 label      : _type_BaseList_init_label       = None,
                 view_max   : _type_BaseList_init_view_max    = 7,
                 focus_color: _type_BaseList_init_focus_color = _colors.basic('cyan'),
                 focus_mark : _type_BaseList_init_focus_mark  = '> ',
                 evade_color: _type_BaseList_init_evade_color = None,
                 evade_mark : _type_BaseList_init_evade_mark  = '  ',
                 fill       : _type_BaseList_init_fill        = True,
                 **kwargs):
        
        self._axis = axis
        
        super_cls = self.__class__.__mro__[1]

        if tiles is _helpers.auto:
            tiles = ()
        
        if not isinstance(tiles, dict):
            tiles = {_get_mesh_spot(axis, index): tile for (index, tile) in enumerate(tiles)}
        
        point = _get_mesh_point(axis, index)

        funnel_enter = _helpers.get_function_arg_safe(super_cls, 'funnel_enter', kwargs, pop = True)
        
        funnel_enter_group = []

        if not axis:
            funnel_enter_entry = _funnels.mesh_flip(axis)
            funnel_enter_group.append(funnel_enter_entry)

        if not view_max is None:
            if not view_max > 1:
                focus_mark = None 
            funnel_enter_entry = _funnels.mesh_max(axis, view_max)
            funnel_enter_group.append(funnel_enter_entry)

        funnel_enter_entry = _funnels.mesh_light(focus_color, evade_color)
        funnel_enter_group.append(funnel_enter_entry)

        if focus_mark is None:
            focus_mark = ''

        if evade_mark is None:
            evade_mark = ''

        if axis == 0:
            funnel_enter_entry = _funnels.mesh_point(focus_mark, evade_mark)
            funnel_enter_group.append(funnel_enter_entry)

        if not label is None:
            def funnel_enter_entry_get(index):
                spot = _get_mesh_spot(axis, index)
                tile = self._mutate.tiles.get(spot)
                value = label(index, tile)
                lines = _helpers.split_lines(value)
                return lines
            funnel_enter_entry = _funnels.mesh_head(axis, funnel_enter_entry_get, _funnels.JustType.start)
            funnel_enter_group.append(funnel_enter_entry)

        if fill:
            funnel_enter_entry = _funnels.mesh_grid_fill()
            funnel_enter_group.append(funnel_enter_entry)

        funnel_enter = _helpers.chain_functions(*funnel_enter_group, funnel_enter)

        super().__init__(
            *args,
            tiles = tiles,
            point = point,
            funnel_enter = funnel_enter,
            **kwargs
        )

    @property
    def axis(self):

        return self._axis

    def _resolve(self):

        index = self._mutate.cur_spot[self._axis]

        return index


_type_Select_init_options = typing.List[str]
_type_Select_init_create  = typing.Union[None, typing.Callable[[_mutates._type_Mesh_init_spot], str]]
_type_Select_init_Option  = Input


class Select(BaseList):

    """
    A single-option selector.

    :param options:
        Converted into :paramref:`.BaseList.tiles`.
    :param create:
        Same as :paramref:`.BaseMesh.create`, except returned values are converted to widgets.
    :param Option:
        Used to convert :paramref:`.options` and :paramref:`.create` results to tiles.

    Arguments directly passed to super-class:
    
        - :paramref:`~.BaseList.tiles` - Created using :paramref:`.options`.
        - :paramref:`~.BaseList.create` - Created using :paramref:`.create`.
        - :paramref:`~.BaseList.focus` - Set to a :class:`~typing.Callable` that delegates no events.

    Arguments used for :paramref:`.Option`:

        - :paramref:`~.Input.value` - Set to each option upon creation.

    |theme| :code:`'widgets.Select'`.
    """

    __slots__ = ()

    @_theme.add('widgets.Select')
    def __init__(self, 
                 *args,
                 options: _type_Select_init_options = _helpers.auto,
                 create : _type_Select_init_create  = None,
                 Option : _type_Select_init_Option  = Input,
                 **kwargs):
        
        def get_tile(value):
            option = Option(
                value = value
            )
            return option
        
        if options is _helpers.auto:
            options = ()
        
        tiles = map(get_tile, options)

        def create(spot, *, __sub = create):
            if __sub is None:
                return None
            value = __sub(spot)
            if not isinstance(value, Widget):
                tile = get_tile(value)
            return tile

        super().__init__(
            *args,
            tiles = tiles, 
            create = create,
            focus = _focus_nil,
            **kwargs
        )


_type_Basket_init_options       = typing.List[str]
_type_Basket_init_active        = typing.List[int]
_type_Basket_init_positive_mark = str
_type_Basket_init_negative_mark = str
_type_Basket_init_Option        = Input
_type_Basket_init_Stamp         = Select


_Basket_focus_events = {
    _core.Event.arrow_left, _core.Event.arrow_right
}

_Basket_focus = lambda event: event in _Basket_focus_events
_Basket_search_get = lambda tile: tile.mutate.tiles[(0, 1)].sketch()


class Basket(BaseList):

    """
    A multi-item selector.

    :param options:
        Converted into :paramref:`.BaseList.tiles`.
    :param active:
        Indexes that should be selected from the start.
    :param positive_mark:
        Prepended to all selected options.
    :param negative_mark:
        Prepended to all non-selected options.
    :param Option:
        Used to convert :paramref:`.options` to tiles.
    :param Stamp:
        Used to convert :paramref:`.positive_mark` or :paramref:`.negative_mark` to tiles.
    
    Arguments directly passed to super-class:
    
        - :paramref:`~.BaseList.tiles` - Created using :paramref:`.options`.
        - :paramref:`~.BaseList.focus` - Set to a :class:`~typing.Callable` that delegates :attr:`~.Event.arrow_left` and :attr:`~.Event.arrow_right`.

    Arguments used for :paramref:`.Option`:

        - :paramref:`~.Input.value` - Set to each option upon creation.

    Arguments used for :paramref:`.Stamp`:

        - :paramref:`~.Select.axis` - Set to :code:`1`.
        - :paramref:`~.Select.options` - Set to a :class:`tuple` of :paramref:`.negative_mark` and :paramref:`.positive_mark`
        - :paramref:`~.BaseList.index` - Set to either :code:`0` or :code:`1` depending on whether the option's index is in :paramref:`.active`.
        - :paramref:`~.BaseList.view_max` - Set to :code:`1`.
        - :paramref:`~.BaseList.focus_color` - Set to :code:`None`.
        - :paramref:`~.Widget.callback` - Set to a :class:`~typing.Callable` that handles multi-switching of marks.

    |theme| :code:`'widgets.Basket'`.
    """

    @_theme.add('widgets.Basket')
    def __init__(self,
                 *args,
                 options      : _type_Basket_init_options       = _helpers.auto, 
                 active       : _type_Basket_init_active        = _helpers.auto,
                 positive_mark: _type_Basket_init_positive_mark = '[X] ', 
                 negative_mark: _type_Basket_init_negative_mark = '[ ] ',
                 Option       : _type_Basket_init_Option        = Input, 
                 Stamp        : _type_Basket_init_Stamp         = Select,
                 **kwargs):
        
        super_cls = self.__class__.__mro__[1]

        if options is _helpers.auto:
            options = ()
        
        if active is _helpers.auto:
            active = set()

        search = _helpers.get_function_arg_safe(super_cls, 'search', kwargs, pop = True)

        if not search is None:
            search = functools.partial(search, get = _Basket_search_get)

        stamp_handle = _handle.Handle()

        def _control_function_arrow_horizontal(forward, info):
            cur_tile = self._mutate.cur_tile
            cur_spot = cur_tile.mutate.cur_tile.mutate.cur_spot
            cur_mark = cur_spot[1]
            if not (forward and cur_mark or not forward and not cur_mark):
                return
            for oth_spot in self._mutate.vision.values():
                oth_tile = self._mutate.tiles[oth_spot]
                if oth_tile is cur_tile:
                    continue
                oth_tile.mutate.cur_tile.mutate.point[1] = cur_mark
            raise _handle.Abort()
        
        for (event, forward) in ((_core.Event.arrow_right, True), (_core.Event.arrow_left, False)):
            partial = functools.partial(_control_function_arrow_horizontal, forward)
            control = _controls.get((_handle.EventType.enter, event))(partial)
            stamp_handle.add(control)

        create_sub = _helpers.get_function_arg_safe(super_cls, 'create', kwargs, pop = True)
        
        def get_widget(index, value):
            stamp_widget_options = (negative_mark, positive_mark)
            stamp_widget_index = int(index in active)
            stamp_widget = Stamp(
                options = stamp_widget_options, 
                axis = 1, 
                index = stamp_widget_index,
                view_max = 1, 
                focus_color = None,
                callback = stamp_handle.invoke
            )
            value_widget = Option(
                value = value
            )
            main_widget_tiles = (stamp_widget, value_widget)
            main_widget = BaseList(
                tiles = main_widget_tiles, 
                axis = 1, 
                focus = True,
                focus_color = None
            )
            return main_widget

        tiles = itertools.starmap(get_widget, enumerate(options))

        def create(index):
            if create_sub is None:
                return None
            value = create_sub(index)
            widget = get_widget(index, value)
            return widget
        
        super().__init__(
            *args,
            tiles  = tiles, 
            create = create, 
            search = search, 
            focus  = _Basket_focus,
            **kwargs)

    def _resolve(self):

        indexes = set()
        for (spot, tile) in self._mutate.tiles.items():
            # row.mutate.stamp.mutate.cur_spot
            if not tile.mutate.cur_tile.mutate.cur_spot[1]:
                continue
            index = spot[self._axis]
            indexes.add(index)
        
        return indexes


_type_Count_init_value   = typing.Union[int, float]
_type_Count_init_rate    = typing.Union[int, float]
_type_Count_init_convert = typing.Union[None, typing.Callable[[_type_Count_init_value], _type_Count_init_value]]
_type_Count_init_decimal = typing.Union[None, bool]
_type_Count_init_Numeric = Numeric
    

_Count_focus_events = {
    _core.Event.insert, _core.Event.delete_left
}

_Count_focus = lambda event: event in _Count_focus_events
    

class Count(BaseList):

    """
    An editable counter.

    Resolves to an :class:`int` (the pointed value).

    :param value:
        Initial value.
    :param rate:
        Rate to move by.
    :param decimal:
        Whether to only allow :class:`float` or :class:`int`.
    :param convert:
        Used with ``(value)`` before creating a new widget.
    :param Numeric:
        Used to convert each value into a widget.

    Arguments directly passed to super-class:
    
        - :paramref:`~.BaseList.axis` - Set to :code:`0`.
        - :paramref:`~.BaseList.index` - Set to :code:`0` (:paramref:`.value` is used as an offset).
        - :paramref:`~.BaseList.clean` - Set to :code:`True`.
        - :paramref:`~.BaseList.create` - Created from :paramref:`.value`, :paramref:`.rate`, :paramref:`.decimal` and :paramref:`.convert`. 
        - :paramref:`~.BaseList.focus` - Set to a :class:`~typing.Callable` that delegates :attr:`~.Event.insert` and :attr:`~.Event.delete_left`.
        - :paramref:`~.BaseList.focus_mark` - Set to a :code:`None`.

    Arguments used for :paramref:`.Numeric`:

        - :paramref:`~.Numeric.value` - Set to each numeric value upon creation.
        - :paramref:`~.Numeric.decimal` - Set to :class:`float` unless :paramref:`.decimal` specifies otherwise.

    |theme| :code:`'widgets.Count'`.
    """

    __slots__ = ('_decimal',)

    @_theme.add('widgets.Count')
    def __init__(self,
                 *args,
                 value  : _type_Count_init_value   = 0,
                 rate   : _type_Count_init_rate    = 1,
                 decimal: _type_Count_init_decimal = None,
                 convert: _type_Count_init_convert = None,
                 Numeric: _type_Count_init_Numeric = Numeric,
                 **kwargs):
        
        self._decimal = decimal
        
        offset = value
        
        def create(spot):
            value = offset - sum(spot) * rate
            decimal = self._decimal
            if decimal is None:
                decimal = True
            if not convert is None:
                value = convert(value) 
            tile = Numeric(
                value = value, 
                decimal = decimal
            )
            return tile
        
        super().__init__(
            *args,
            axis = 0,
            index = 0,
            clean = True,
            create = create,
            focus = _Count_focus,
            focus_mark = None,
            **kwargs
        )

    def _resolve(self):

        value_any = self._mutate.cur_tile.resolve()
        value_int = int(value_any)

        if self._decimal is None and value_int == value_any:
            value_any = value_int

        return value_any
    

_DateTime_focus_events = {
    _core.Event.arrow_up, _core.Event.arrow_down,
    _core.Event.insert, _core.Event.delete_left
}


_DateTime_focus = lambda event: event in _DateTime_focus_events


_DateTime_funnel_enter_arrange_attr_groups = (
    ('year', 'month', 'day'),
    ('hour', 'minute', 'second')
)


def _DateTime_funnel_enter_arrange(axis, attrs, date_delimit, time_delimit, part_delimit, tiles, point):

    delimits = (date_delimit, time_delimit)

    stores = ([], [])

    for attr in attrs:
        for (attr_group_index, attr_group) in enumerate(_DateTime_funnel_enter_arrange_attr_groups):
            if not attr in attr_group:
                continue
            break
        else:
            raise ValueError(f'invalid attribute: {attr}')
        stores[attr_group_index].append(attr)

    index = 0
    last_tiles = {}
    roll_tiles = {}

    for some_attrs, some_delimit in zip(stores, delimits):
        if not some_attrs:
            continue
        some_tiles = {}
        for index in range(index, len(some_attrs) + index):
            spot = _get_mesh_spot(axis, index)
            tile = tiles[spot]
            some_tiles[spot] = tile
        _funnels.mesh_delimit.call(axis, some_delimit, some_tiles, None)
        try:
            max_spot = max(last_tiles)
            min_spot = min(some_tiles)
        except ValueError:
            pass
        else:
            roll_tiles[max_spot] = last_tiles[max_spot]
            roll_tiles[min_spot] = some_tiles[min_spot]
        index += 1
        last_tiles = some_tiles

    if not roll_tiles:
        return
    
    _funnels.mesh_delimit.call(axis, part_delimit, roll_tiles, None)


_Datetime_zfills = {
    'year'  : 4,
    'month' : 2,
    'day'   : 2,
    'hour'  : 2,
    'minute': 2,
    'second': 2
}


_type_DateTime_init_value        = datetime.datetime
_type_DateTime_init_Chron        = Count
_type_DateTime_init_attrs        = typing.List[str]
_type_DateTime_init_date_delimit = str
_type_DateTime_init_time_delimit = str
_type_DateTime_init_part_delimit = str


class DateTime(BaseList):

    """
    A datetime picker.

    Resolves to :class:`datetime.datetime` object.

    :param value:
        Initial value.
    :param attrs:
        The :class:`datetime` attributes used.
    :param Chron:
        Used to convert :paramref:`.value` to tiles.
    :param date_delimit:
        Separator for date parts.
    :param time_delimit:
        Separator for time parts.
    :param part_delimit:
        Separator for the date and time parts.

    Arguments directly passed to super-class:
    
        - :paramref:`~.BaseList.axis` - Set to :code:`1`.
        - :paramref:`~.BaseList.tiles` - Created from :paramref:`.value`'s attributes using :paramref:`.attrs` and :paramref:`.Chron`.
        - :paramref:`~.BaseMesh.focus` - Set to a :class:`~typing.Callable` that delegates :attr:`~.Event.arrow_up`, :attr:`~.Event.arrow_down`, :attr:`~.Event.insert` and :attr:`~.Event.delete_left`. 

    Arguments used for :paramref:`.Chron`:

        - :paramref:`~.Count.value` - Set to each numeric value for each part.
        - :paramref:`~.Count.decimal` - Set to :code:`False`.
        - :paramref:`~.Count.convert` - Set to a :class:`~typing.Callable` that returns the rolling :class:`datetime.datetime`\'s value after attempting to set it.
        - :paramref:`~.Count.Numeric` - Set to a partial :class:`.Numeric` (see later for arguments).
        - :paramref:`~.BaseList.focus_color` - Set to :code:`None`.

    Arguments used for :paramref:`.Chron`\'s :paramref:`~.Count.Numeric`:

        - :paramref:`~.Numeric.zfill` - Set to the maximum amount of digits for the respective part.

    |theme| :code:`'widgets.DateTime'`.
    """

    __slots__ = ('_datetime', '_datetime_attrs')

    @_theme.add('widgets.DateTime')
    def __init__(self,
                 *args,
                 value       : _type_DateTime_init_value        = None,
                 attrs       : _type_DateTime_init_attrs        = ('day', 'month', 'year', 'hour', 'minute'),
                 Chron       : _type_DateTime_init_Chron        = Count,
                 date_delimit: _type_DateTime_init_date_delimit = '/',
                 time_delimit: _type_DateTime_init_time_delimit = ':',
                 part_delimit: _type_DateTime_init_part_delimit = ' ',
                 **kwargs):
        
        super_cls = self.__class__.__mro__[1]
        
        funnel_enter = _helpers.get_function_arg_safe(super_cls, 'funnel_enter', kwargs, pop = True)

        axis = 1
        
        if value is None:
            value = datetime.datetime.now()

        self._datetime = value

        self._datetime_attrs = attrs

        def get_tile(attr):
            def convert(value):
                kwargs = {attr: value}
                self._datetime = self._convert(kwargs)
                return str(value)
            zfill = _Datetime_zfills.get(attr, 0)
            chron = Chron(
                value = getattr(value, attr),
                decimal = False,
                convert = convert,
                Numeric = functools.partial(
                    Numeric, 
                    zfill = zfill
                ),
                focus_color = None
            )
            return chron
        
        tiles = []
        for attr in attrs:
            tile = get_tile(attr)
            tiles.append(tile)

        funnel_enter_group = []

        funnel_enter_entry = functools.partial(
            _DateTime_funnel_enter_arrange,
            axis, attrs, date_delimit, time_delimit, part_delimit,
        )
        funnel_enter_group.append(funnel_enter_entry)

        funnel_enter = _helpers.chain_functions(funnel_enter, *funnel_enter_group)

        super().__init__(
            *args,
            axis = 1,
            tiles = tiles,
            focus = _DateTime_focus,
            funnel_enter = funnel_enter,
            **kwargs
        )

    @property
    def attrs(self):

        return self._datetime_attrs
    
    def _convert(self, kwargs, value = None):

        try:
            datetime = self._datetime.replace(**kwargs)
        except ValueError as error: 
            message = str(error)
            if not value is None:
                message += f' ({value})'
            raise Abort(message)
        
        return datetime

    def _resolve(self):

        kwargs = {}
        for (index, name) in enumerate(self._datetime_attrs):
            spot = _get_mesh_spot(self._axis, index)
            tile = self._mutate.tiles[spot]
            value = tile.resolve()
            kwargs[name] = value

        value = self._convert(kwargs)

        return value
    

_Form_tile_focus = lambda event: True
    

class Form(BaseList):

    """
    A multi-item form.

    :param form:
        A mapping of ``field: editable widget`` pairs.
    :param Field:
        Used to convert :paramref:`.form` keys to tiles.
    :param delimit:
        Placed between fields and widgets.

    Arguments directly passed to super-class:
    
        - :paramref:`~.BaseList.axis` - Set to :code:`1`
        - :paramref:`~.BaseList.tiles` - Created using :paramref:`.form`.
        - :paramref:`~.BaseList.focus_color` - Set to :code:`None`.
        - :paramref:`~.BaseList.evade_color` - Set to :code:`None`.

    Arguments used for :paramref:`.Field`:

        - :paramref:`~.Input.value` - Set to each pair's field.
        - :paramref:`~.Input.funnel_leave` - Set to a :class:`~typing.Callable` that aligns with others fields and includes :paramref:`.delimit`.

    |theme| :code:`'widgets.Form'`.
    """

    @_theme.add('widgets.Form')
    def __init__(self, 
                 *args, 
                 form = {},
                 Field = Input, 
                 delimit = ': ',
                 **kwargs):
        
        super_cls = self.__class__.__mro__[1]
        
        form = dict(form)

        def tile_focus(*args):
            return _Form_tile_focus(*args)
        
        focus_color = _helpers.get_function_arg_safe(super_cls, 'focus_color', kwargs, pop = True)
        evade_color = _helpers.get_function_arg_safe(super_cls, 'evade_color', kwargs, pop = True)
                    
        def get_tile_focus_color(index):
            def focus_color_get():
                if not self._mutate.cur_spot[0] == index:
                    return None
                return focus_color
            return focus_color_get
        
        def get_tile_evade_color(index):
            def evade_color_get():
                if not self._mutate.cur_spot[0] == index:
                    return None
                return evade_color
            return evade_color_get
        
        top_field_size = max(map(len, form))

        def get_tile(index, field, value_widget):
            field_funnel_leave_group = []
            field_funnel_leave_entry = _funnels.text_min_horizontal(_funnels.JustType.end, top_field_size, ' ')
            field_funnel_leave_group.append(field_funnel_leave_entry)
            field_funnel_leave_entry = _funnels.text_bloat_horizontal(_funnels.JustType.start, 1, delimit)
            field_funnel_leave_group.append(field_funnel_leave_entry)
            field_funnel_leave = _helpers.chain_functions(*field_funnel_leave_group)
            field_widget = Field(
                value = field,
                funnel_leave = field_funnel_leave
            )
            tile_widgets = (field_widget, value_widget)
            tile = BaseList(
                axis = 1,
                index = 1,
                tiles = tile_widgets,
                focus = tile_focus,
                focus_color = get_tile_evade_color(index),
                evade_color = get_tile_focus_color(index)
            )
            return tile
        
        tiles = []
        for index, item in enumerate(form.items()):
            tile = get_tile(index, *item)
            tiles.append(tile)

        super().__init__(
            *args,
            axis = 0,
            tiles = tiles,
            focus_color = None,
            evade_color = None,
            **kwargs
        )

    def _resolve(self):

        form = {}
        for tile in self.mutate.tiles.values():
            field_widget = tile.mutate.tiles[(0, 0)]
            field = field_widget.resolve()
            value_widget = tile.mutate.tiles[(0, 1)]
            value = value_widget.resolve()
            form[field] = value

        return form



import contextlib
import importlib
import os
import threading
import copy
import functools


def noop(*args, **kwargs):

    pass


@contextlib.contextmanager
def noop_contextmanager(*args, **kawrgs):

    yield


def import_module_os(name):

    name = '.' + name + '_os_' + os.name

    module = importlib.import_module(name, package = __package__)

    return module


def ctxmethod(get):

    def decorator(function):
        @functools.wraps(function)
        def wrapper(self, *args, **kwargs):
            with get(self):
                return function(self, *args, **kwargs)
        return wrapper
    
    return decorator


class Atomic:

    __slots__ = ('__enter', '__leave', '_size', '_lock')

    def __init__(self, enter = None, leave = None):

        self.__enter = enter
        self.__leave = leave

        self._size = 0
        self._lock = threading.Lock()

    @property
    def open(self):

        return not self._size

    @ctxmethod(lambda self: self._lock)
    def _deduce(self, function, step, limit):

        size = self._size + step

        if not size > limit and function:
            function()

        self._size = size

    def _enter(self):

        self._deduce(self.__enter, 1, 1)

    def enter(self):

        self._enter()

    def _leave(self):

        self._deduce(self.__leave, - 1, 0)

    def leave(self):

        self._leave()

    def __enter__(self):

        self._enter()

    def __exit__(self, type, value, traceback):

        self._leave()




from . import _ansi as ansi

from ._io     import IO
from ._intel  import Intel
from ._cursor import Cursor, ClearMode, EraseMode
from ._source import Source, Event
from ._handle import Handle, Terminate
from ._render import Render
from ._screen import Screen

from ._console import Console, SkipDraw

from ._ansi   import _type_parse_return as _type_ansi_parse_return
from ._render import _type_Render_draw_lines, _type_Render_draw_point
from ._screen import _type_Screen_draw_sketch
from ._handle import _type_Handle_start_invoke


__all__ = (
    'ansi', 
    'IO', 'Intel', 'Cursor', 'ClearMode', 'EraseMode', 'Source', 'Event', 'Render', 'Screen', 'Handle', 'Terminate', 
    'Console', 'SkipDraw'
)



"""
Classes for transforming specific-structure data into a form that be 
used by to print to the console while maintaining proper cursor position.

Most accept data structured in similar ways to how it is used in :class`~.mutate.Mutate`\ s.
"""

import typing
import abc
import copy
import math

from . import _core


__all__ = ('Visual', 'Text', 'Mesh', 'Line')


_type_Visual_init_get          = typing.Callable[[bool, bool], typing.Tuple[typing.Any]]
_type_Visual_init_funnel_enter = typing.Union[typing.Callable[[typing.Any], None], None]
_type_Visual_init_funnel_leave = typing.Union[typing.Callable[[_core._type_Render_draw_lines, _core._type_Render_draw_point], None], None]

_type_Visual_get_enter         = bool
_type_Visual_get_leave         = bool
_type_Visual_get_return        = typing.Union[typing.Tuple[_core._type_Render_draw_lines, _core._type_Render_draw_point] , None]


class Visual(abc.ABC):

    """
    Base class for transformers of data into lines and point for rendering.

    :param get:
        Used for fetching the data.
        The arguments have the same meaning as in :meth:`.get`.
    :param funnel_enter:
        Used for mutating the data in-place before turning into lines and point.
    :param funnel_leave:
        Used for mutating the resulting lines and point in-place after transforming the data.
    """

    __slots__ = ('_get', '_funnel_enter', '_funnel_leave')

    def __init__(self, 
                 get         : _type_Visual_init_get, 
                 funnel_enter: _type_Visual_init_funnel_enter = None,
                 funnel_leave: _type_Visual_init_funnel_leave = None):

        self._get = get

        self._funnel_enter = funnel_enter
        self._funnel_leave = funnel_leave

    @abc.abstractmethod
    def _format(self, *args):

        return NotImplemented

    def get(self, 
            enter: _type_Visual_get_enter = True, 
            leave: _type_Visual_get_leave = True) -> _type_Visual_get_return:

        """
        Fetches and transforms the data.

        :param enter:
            Whether to use ``funnel_enter``.
        :param leave:
            Whether to use ``funnel_leave``.
        
        :return:
            The final lines and point.
        """

        assets = self._get(enter, leave)

        assets = copy.deepcopy(assets)

        if enter and not self._funnel_enter is None:
            self._funnel_enter(*assets)

        (lines, point) = self._format(*assets)

        if leave and not self._funnel_leave is None:
            self._funnel_leave(lines, point)

        return (lines, point)


_type_Text_link_lines        = typing.List[typing.List[str]]
_type_Text_link_point        = typing.List[int]

_type_Text_init_get          = typing.Callable[[bool, bool], typing.Tuple[_type_Text_link_lines, _type_Text_link_point]]
_type_Text_init_funnel_enter = typing.Union[typing.Callable[[_type_Text_link_lines, _type_Text_link_point], None], None]
_type_Text_init_funnel_leave = _type_Visual_init_funnel_leave


class Text(Visual):

    """
    Transforms lines and point into... lines and cursor point.

    :param get:
        Used for fetching ``data``.
    :param funnel_enter:
        Used for mutating ``data`` in-place before turning into ``(lines, point)``.
    :param funnel_leave:
        Used for mutating the resulting ``(lines, point)`` in-place after transforming ``data``.
    """

    __slots__ = ()

    def __init__(self, 
                 get         : _type_Text_init_get, 
                 funnel_enter: _type_Text_init_funnel_enter = None,
                 funnel_leave: _type_Text_init_funnel_leave = None):
        
        super().__init__(get, funnel_enter, funnel_leave)

    @classmethod
    def link(cls, 
             lines: _type_Text_link_lines, 
             point: _type_Text_link_point, 
             *args, **kwargs):

        """
        Use this if ``(lines, point)`` are expected to be mutated in-place over their life time.
        """

        def get(*args):
            return (lines, point)

        return cls(get, *args, **kwargs)

    def _format(self, lines, point):

        return (lines, point)
    

_type_Mesh_link_tiles        = typing.Dict[typing.Tuple[int, int], typing.Tuple[_type_Text_link_lines, _type_Text_link_point]]
_type_Mesh_link_point        = typing.List[int]

_type_Mesh_init_get          = typing.Callable[[bool, bool], typing.Tuple[_type_Mesh_link_tiles, _type_Mesh_link_point]]
_type_Mesh_init_funnel_enter = typing.Union[typing.Callable[[_type_Mesh_link_tiles, _type_Mesh_link_point], None], None]
_type_Mesh_init_funnel_leave = _type_Visual_init_funnel_leave


class Mesh(Visual):

    """
    Transforms mesh tiles and point into lines and cursor point.

    :param get:
        Used for fetching ``(tiles, point)``.
    :param funnel_enter:
        Used for mutating ``(tiles, point)`` in-place before turning into ``(lines, point)``.
    :param funnel_leave:
        Used for mutating the resulting ``(lines, point)`` in-place after transforming ``(tiles, point)``.

    - ``tiles`` is ``{spot: tile, ...}``
    - ``spot`` is ``(y, x)``
    - ``tile`` is ``(lines, point)``
    """

    __slots__ = ()

    def __init__(self, 
                 get         : _type_Mesh_init_get, 
                 funnel_enter: _type_Mesh_init_funnel_enter = None,
                 funnel_leave: _type_Mesh_init_funnel_leave = None):
        
        super().__init__(get, funnel_enter, funnel_leave)

    @classmethod
    def link(cls, 
             tiles: _type_Mesh_link_tiles, 
             point: _type_Mesh_link_point, 
             *args, **kwargs):

        """
        Use this if ``(tiles, point)`` are expected to be mutated in-place over their life time.

        - ``tiles`` is ``{spot: get, ...}``
        - ``spot`` is ``(y, x)``
        - ``get`` is like :meth:`.get` and returns ``(lines, point)``
        """

        def get(*args):
            dyn_tiles = {spot: get(*args) for (spot, get) in tiles.items()}
            dyn_point = copy.deepcopy(point)
            return (dyn_tiles, dyn_point)
        
        return cls(get, *args, **kwargs)

    def _format(self, tiles, point):

        spots = tiles.keys()

        (main_y, main_x) = point

        done_lines = []
        memo_lines = []
        memo_y = - math.inf

        fin_y = 0
        fin_x = 0

        for spot in sorted(spots):
            (spot_y, spot_x) = spot
            (tile_lines, tile_point) = tiles[spot]
            (tile_point_y, tile_point_x) = tile_point
            if not spot_y == memo_y:
                if spot_y <= main_y:
                    fin_y += len(memo_lines)
                done_lines[0:0] = memo_lines
                memo_y = spot_y
                memo_lines.clear()
            if spot_y == main_y and spot_x == main_x:
                fin_y += len(tile_lines) - tile_point_y
                try:
                    memo_line = memo_lines[tile_point_y]
                except IndexError:
                    ext_x = 0
                else:
                    ext_x = len(memo_line)
                fin_x = ext_x + tile_point_x
            for (tile_line_index, tile_line) in enumerate(tile_lines):
                try:
                    memo_line = memo_lines[tile_line_index]
                except IndexError:
                    memo_line = []
                    memo_lines.append(memo_line)
                memo_line.extend(tile_line)

        done_lines[0:0] = memo_lines

        fin_y = len(done_lines) - fin_y

        done_point = [fin_y, fin_x]

        return (done_lines, done_point)


_type_Line_link_tiles        = typing.Dict[typing.Tuple[int, int], typing.Tuple[_type_Text_link_lines, _type_Mesh_link_point]]
_type_Line_link_point        = typing.List[int]

_type_Line_init_get          = typing.Callable[[bool, bool], typing.Tuple[_type_Line_link_tiles, _type_Line_link_point]]
_type_Line_init_funnel_enter = typing.Union[typing.Callable[[_type_Line_link_tiles, _type_Line_link_point], None], None]
_type_Line_init_funnel_leave = _type_Visual_init_funnel_leave


class Line(Visual):

    """
    Transforms a list of tiles and point into lines and cursor point.

    :param get:
        Used for fetching ``(tiles, point)``.
    :param funnel_enter:
        Used for mutating ``(tiles, point)`` in-place before turning into ``(lines, point)``.
    :param funnel_leave:
        Used for mutating the resulting ``(lines, point)`` in-place after transforming ``(tiles, point)``.

    - ``tiles`` is ``[get, ...]``
    - ``get`` is like :meth:`.get` and return ``(lines, point)``
    """

    __slots__ = ()

    def __init__(self, 
                 get         : _type_Line_init_get, 
                 funnel_enter: _type_Line_init_funnel_enter = None,
                 funnel_leave: _type_Line_init_funnel_leave = None):
        
        super().__init__(get, funnel_enter, funnel_leave)

    @classmethod
    def link(cls, 
             tiles: _type_Line_link_tiles, 
             point: _type_Line_link_point, 
             *args, **kwargs):

        """
        Use this if ``(tiles, point)`` are expected to be mutated in-place over their life time.

        - ``tiles`` is ``[get, ...]``
        - ``get`` is like :meth:`.get`
        """

        def get(*args):
            dyn_tiles = [get(*args) for get in tiles]
            dyn_point = copy.deepcopy(point)
            return (dyn_tiles, dyn_point)
        
        return cls(get, *args, **kwargs)

    def _format(self, tiles, tiles_point):

        tiles_index = tiles_point[0]

        fin_y = 0
        fin_x = 0

        lines = [[]]
        for (tile_index, (tile_lines, tile_point)) in enumerate(tiles):
            if tile_index < tiles_index:
                fin_y += len(tile_lines) - 1
            elif tile_index == tiles_index:
                fin_y += tile_point[0]
                fin_x = tile_point[1]
                if not fin_y:
                    fin_x += len(lines[fin_y])
            try:
                line = tile_lines.pop(0)
            except IndexError:
                continue
            lines[- 1].extend(line)
            lines.extend(tile_lines)

        point = [fin_y, fin_x]

        return (lines, point)




from . import _core as core

from . import _system as system
from . import _colors as colors

from . import _handle as handle
from . import _mutates as mutates
from . import _visuals as visuals

from . import _theme as theme
from . import _utils as utils

from . import _funnels as funnels
from . import _widgets as widgets

from . import _routines as routines
from . import _printers as printers
from . import _graphics as graphics


__all__ = (
    'system', 'colors', 'funnels', 'theme', 'utils', 
    'widgets', 'latent', 'routines', 'printers', 'graphics'
)


import typing
import collections


__all__ = ('Text', 'Control', 'Escape', 'Sequence', 'parse', 'parse_iterable', 'get_escape', 'get_control')


Text = collections.namedtuple('Text', 'rune')
"""
Non-ANSI text.
"""

Control = collections.namedtuple('Control', 'rune')
"""
ANSI control code (`wiki <https://en.wikipedia.org/wiki/ANSI_escape_code#C0_control_codes>`__).
"""

Escape = collections.namedtuple('Escape', 'rune')
"""
ANSI escape sequence (`wiki <https://en.wikipedia.org/wiki/ANSI_escape_code#Fe_Escape_sequences>`__).
"""

Sequence = collections.namedtuple('Sequence', 'rune args trail')
"""
ANSI control sequence (`wiki <https://en.wikipedia.org/wiki/ANSI_escape_code#CSI_(Control_Sequence_Introducer)_sequences>`__).
"""


_BITS_CODE_C0_RANGE = {*range(0,  32), 127}

_BITS_ESCAPE = 27

_BITS_CODE_C1_RANGE = {*range(64, 96)}

_BITS_SEQUENCE_INTRODUCER = 91

_BITS_SEQUENCE_PARAM_ENTER_RANGE = {*range(48, 64)}

_BITS_SEQUENCE_PARAM_SEPARATOR = 59

_BITS_SEQUENCE_PARAM_LEAVE_RANGE = {*range(32, 48)}

_BITS_SEQUENCE_FINAL_RANGE = {*range(64, 127)}


def _parse_sequence(get):

    def join(store):
        return ''.join(store)

    def update(buffer, store):
        store.append(join(buffer))
        buffer.clear()

    buffer = []
    params = []
    while True:
        rune = get()
        bits = ord(rune)
        if bits == _BITS_SEQUENCE_PARAM_SEPARATOR:
            if not buffer:
                buffer.append('0')
            update(buffer, params)
        elif bits in _BITS_SEQUENCE_PARAM_ENTER_RANGE:
            buffer.append(rune)
        else:
            break

    if buffer:
        update(buffer, params)

    trail = []
    while True:
        if bits in _BITS_SEQUENCE_PARAM_LEAVE_RANGE:
            buffer.append(rune)
        else:
            break
        rune = get()
        bits = ord(rune)

    trail = join(trail)

    # if not bits in _BITS_SEQUENCE_FINAL_RANGE:
    #     raise RuntimeError(
    #         'unexpected sequence rune: {0}'.format(rune)
    #     )

    params = tuple(params)

    return Sequence(rune, params, trail)


def _parse_escape(get):

    rune = get()
    bits = ord(rune)

    if bits == _BITS_SEQUENCE_INTRODUCER:
        return _parse_sequence(get)

    # if len(rune) and not bits in _BITS_CODE_C1_RANGE:
    #     raise RuntimeError(
    #         'unexpected escape rune: {0}'.format(rune)
    #     )

    return Escape(rune)


def _parse_stream(get):

    memo = []

    def function():
        for _ in range(2):
            try:
                rune = memo.pop(0)
            except IndexError:
                text = get()
                memo.extend(text)
            else:
                break
        try:
            rune
        except NameError:
            rune = ''
        return rune

    return function


def _parse(get):

    get = _parse_stream(get)

    rune = get()
    bits = ord(rune)

    if bits in _BITS_CODE_C0_RANGE:
        if bits == _BITS_ESCAPE:
            return _parse_escape(get)
        return Control(rune)

    return Text(rune)


_type_parse_return = typing.Union[Text, Escape, Control, Sequence]


def parse(get: typing.Callable[[], str]) -> _type_parse_return:

    """
    Parse an incoming series of characters into their ANSI representation.
     
    :param get:
        Used whenever an additional character is needed, may return an empty string.

    :return:
        Encapsulates resulting ANSI information.
    """

    return _parse(get)


def _parse_iterable(iterable):

    iterator = iter(iterable)

    return _parse(iterator.__next__)


def parse_iterable(iterable: typing.Iterable) -> _type_parse_return:

    """
    Same as :func:`.parse`, except accepts an iterable.

    :param iterable:
        Its :code:`iter(...).__next__` method is used as :paramref:`.parse.get`.

    :return:
        Same as in :func:`.parse`.
    """

    return _parse_iterable(iterable)


_TEXT_ESCAPE = chr(_BITS_ESCAPE)

_TEXT_SEQUENCE_INTRODUCER = chr(_BITS_SEQUENCE_INTRODUCER)

_TEXT_SEQUENCE_INTRODUCER_CONTROL = _TEXT_ESCAPE + _TEXT_SEQUENCE_INTRODUCER

_TEXT_SEQUENCE_PARAM_SEPARATOR = chr(_BITS_SEQUENCE_PARAM_SEPARATOR)

_TEXT_SEQUENCE_PRIVATE_FLAG = '?'


def _get(prefix, code, *args):

    args = ('' if arg is None else str(arg) for arg in args)

    body = _TEXT_SEQUENCE_PARAM_SEPARATOR.join(args)

    return prefix + body + code


def _get_escape(*args):

    code = _TEXT_ESCAPE

    return _get(code, *args)


def get_escape(code: str, *args) -> str:

    """
    Get the escape sequence of ``code`` with ``args``.
    """

    return _get_escape(code, *args)


def _get_control(private, *args):

    flag = _TEXT_SEQUENCE_PRIVATE_FLAG if private else ''

    code = _TEXT_SEQUENCE_INTRODUCER_CONTROL + flag

    return _get(code, *args)


def get_control(code: str, *args, private = False) -> str:

    """
    Get the control sequence of ``code`` with ``args``.
    """

    return _get_control(private, code, *args)




"""
Functions that can be used directly for :paramref:`.mutates.Mesh.score`.
"""

import itertools


__all__ = ('fuzzy',)


_fuzzy_mark = object()


def fuzzy(argument, tile, get = lambda tile: tile.sketch(False, False)):

    (lines, point) = get(tile)

    cur_line = itertools.chain.from_iterable(lines)
    cur_line = map(str.lower, cur_line)
    cur_line = list(cur_line)
    may_line = itertools.chain.from_iterable(argument)
    may_line = map(str.lower, may_line)

    score = 0
    density = 0
    for (may_index, may_rune) in enumerate(may_line):
        try:
            cur_index = cur_line.index(may_rune)
        except ValueError:
            return None
        score += 1
        density -= abs(may_index - cur_index)
        cur_line[cur_index] = _fuzzy_mark

    return (score, density)



"""
Functions that use :class:`~widget.Widget`\ s within a :class:`.stage.get` 
visual and respond with the formated result, before returning it upon submission.
"""

import os

from . import _helpers
from . import _colors
from . import _widgets
from . import _theme


__all__ = ('input', 'conceal', 'numeric', 'inquire', 'select', 'basket', 'datetime', 'form')


@_theme.add('routines.input.reply')
def _input_reply(widget,
                 result,
                 color = _colors.basic('cyan')):
    
    """
    |theme| :code:`'routines.input.reply'`
    """
    
    result = _helpers.paint_text(color, result)

    return result


@_theme.add('routines.input')
def _input(*args, 
           reply = _input_reply,
           **kwargs):
    
    """
    Use an :class:`~.widgets.Input` widget.

    All widget and :func:`~.widgets.start` arguments are valid.

    |theme| :code:`'routines.input'`

    .. image:: _static/images/routines.input-1.gif

    .. code-block:: python

        value = survey.routines.input('ping? ')
        print(f'Answered {value}.')

    .. image:: _static/images/routines.input-2.gif

    .. code-block:: python

        limit = 50
        def info(widget, name, info):
            result = widget.resolve().rstrip('\\n')
            remain = limit - len(result)
            if remain < 0:
                raise survey.widgets.Abort('no characters left')
            return str(remain)
        value = survey.routines.input('Limited Message: ', multi = True, info = info)
        print(f'Answered with {len(value)} characters.')
    """

    widget_kwargs_names = _helpers.get_function_args_names(_widgets.Input)
    widget_kwargs = _helpers.yank_dict(kwargs, widget_kwargs_names)
    widget = _widgets.Input(**widget_kwargs)

    multi_pre = widget_kwargs.get('multi', False)
    multi_aft = multi_pre

    result = _widgets.start(
        multi_pre, 
        multi_aft, 
        widget, 
        *args, 
        reply = reply, 
        **kwargs
    )

    return result


input = _input


@_theme.add('routines.numeric.reply')
def _numeric_reply(widget, 
                   result,
                   color = _colors.basic('cyan')):
    
    result = str(result)
    
    result = _helpers.paint_text(color, result)

    return result


@_theme.add('routines.numeric')
def _numeric(*args,
             reply = _numeric_reply,
             **kwargs):
    
    """
    Use an :class:`~.widgets.Count` widget.

    All widget and :func:`~.widgets.start` arguments are valid.

    |theme| :code:`'routines.numeric'`

    .. image:: _static/images/routines.numeric-1.gif

    .. code-block:: python

        value = survey.routines.numeric('Price: ')
        print(f'Answered {value}.')

    .. image:: _static/images/routines.numeric-2.gif

    .. code-block:: python

        value = survey.routines.numeric('Age: ', decimal = False)
        print(f'Answered {value}.')
    """

    widget_kwargs_names = _helpers.get_function_args_names(_widgets.Count)
    widget_kwargs = _helpers.yank_dict(kwargs, widget_kwargs_names)
    widget = _widgets.Count(**widget_kwargs)

    multi_pre = False
    multi_aft = multi_pre

    result = _widgets.start(
        multi_pre, 
        multi_aft, 
        widget, 
        *args, 
        reply = reply, 
        **kwargs
    )

    return result


numeric = _numeric


@_theme.add('routines.conceal.reply')
def _conceal_reply(widget, 
                   result, 
                   rune = '*', 
                   color = _colors.basic('black')):

    result = rune * len(result)

    result = _helpers.paint_text(color, result)

    return result


@_theme.add('routines.conceal')
def _conceal(*args,
             reply = _conceal_reply, 
             **kwargs):
    
    """
    Use an :class:`~.widgets.Conceal` widget.

    All widget and :func:`~.widgets.start` arguments are valid.

    |theme| :code:`'routines.conceal'`

    .. image:: _static/images/routines.conceal-1.gif

    .. code-block:: python
        
        value = survey.routines.conceal('Password: ')
        print(f'Answered {value}.')
    """

    widget_kwargs_names = _helpers.get_function_args_names(_widgets.Conceal)
    widget_kwargs = _helpers.yank_dict(kwargs, widget_kwargs_names)
    widget = _widgets.Conceal(**widget_kwargs)

    multi_pre = False
    multi_aft = multi_pre

    result = _widgets.start(
        multi_pre, 
        multi_aft, 
        widget, 
        *args, 
        reply = reply, 
        **kwargs
    )

    return result


conceal = _conceal


@_theme.add('routines.inquire.reply')
def _inquire_reply(widget, 
                   result, 
                   color = _colors.basic('cyan')):

    result = 'Yes' if result else 'No'

    result = _helpers.paint_text(color, result)

    return result


@_theme.add('routines.inquire')
def _inquire(*args, 
             default_color = _colors.basic('cyan'), 
             reply = _inquire_reply,
             **kwargs):
    
    """
    Use an :class:`~.widgets.Inquire` widget.

    :param options:
        The keys are used as :paramref:`.widgets.Inquire.options`. The resolving key's value is returned.
    :param default_color:
        Used in the default hint for signifying the default value. The default value must be part of :paramref:`.options`' values.

    All other widget and :func:`~.widgets.start` arguments are valid.

    |theme| :code:`'routines.inquire'`

    .. image:: _static/images/routines.inquire-1.gif

    .. code-block:: python

        value = survey.routines.inquire('Save? ', default = True)
        print(f'Answered {value}.')
    """

    widget_kwargs_names = _helpers.get_function_args_names(_widgets.Inquire)
    widget_kwargs = _helpers.yank_dict(kwargs, widget_kwargs_names)
    
    options = _helpers.get_function_arg_safe(_widgets.Inquire, 'options', widget_kwargs)
    default_value = _helpers.get_function_arg_safe(_widgets.Inquire, 'default', widget_kwargs)
    default_option = next((option for (option, value) in options.items() if value == default_value), default_value)

    widget = _widgets.Inquire(**widget_kwargs)

    if not 'hint' in kwargs:
        hint_transform = lambda value: _helpers.paint_text(default_color, value.title())
        hint_segments = (hint_transform(option) if option == default_option else option for option in options)
        hint = kwargs['hint'] = '(' + '/'.join(hint_segments) + ')' + ' '

    multi_pre = False
    multi_aft = multi_pre

    value = _widgets.start(
        multi_pre, 
        multi_aft, 
        widget, 
        *args, 
        reply = reply, 
        **kwargs
    )

    return value


inquire = _inquire


def _select_hint_gen_instructions(widget):

    if not widget.mutate._search_score is None:
        yield 'filter: type'

    yield 'move: ↑↓'


def _select_hint(widget, name, info):

    instructions = _select_hint_gen_instructions(widget)

    value = '[' + ' | '.join(instructions) + ']'

    return value


@_theme.add('routines.select.reply')
def _select_reply(widget, 
                  index,
                  color = _colors.basic('cyan')):

    value = widget.mutate.cur_tile.resolve()
    
    value = _helpers.paint_text(color, value)

    return value


@_theme.add('routines.select')
def _select(*args, 
            reply = _select_reply, 
            **kwargs):
    
    """
    Use an :class:`~.widgets.Select` widget.

    The default hint shows available controls.

    All widget and :func:`~.widgets.start` arguments are valid.

    |theme| :code:`'routines.select'`

    .. image:: _static/images/routines.select-1.gif

    .. code-block:: python

        colors = ('red', 'green', 'blue', 'pink', 'silver', 'magenta')
        index = survey.routines.select('Pick a color: ', options = colors)
        print(f'Answered {index}.')

    .. image:: _static/images/routines.select-2.gif    

    .. code-block:: python

        names = map(''.join, itertools.combinations('AXBYCZ', 3))
        index = survey.routines.select('Favorite names? ',  options = names,  focus_mark = '~ ',  evade_color = survey.colors.basic('yellow'))
        print(f'Answered {index}.')
    """

    widget_kwargs_names = _helpers.get_function_args_names(_widgets.Select)
    widget_kwargs = _helpers.yank_dict(kwargs, widget_kwargs_names)
    widget = _widgets.Select(**widget_kwargs)

    kwargs.setdefault('hint', _select_hint)
    kwargs.setdefault('hint_parse', True)

    def info(widget, *args):
        mutate = widget.mutate
        lines = mutate.search_lines
        point = mutate.search_point
        return (lines, point)

    multi_pre = True
    multi_aft = False

    result = _widgets.start(
        multi_pre,
        multi_aft,
        widget,
        *args,
        info = info,
        info_parse = False,
        site = 'info',
        reply = reply,
        **kwargs
    )

    return result


select = _select


def _basket_hint_gen_instructions(widget):

    if not widget.mutate._search_score is None:
        yield 'filter: type'

    yield 'move: ↑↓'

    yield 'pick: → all: →→'
    yield 'unpick: ← all: ←←'


def _basket_hint(widget, name, info):

    instructions = _basket_hint_gen_instructions(widget)

    value = '[' + ' | '.join(instructions) + ']'

    return value


@_theme.add('routines.basket.reply')
def _basket_reply(widget, 
                  indexes, 
                  color = _colors.basic('cyan'), 
                  delimiter = ', '):
    
    size = len(widget.mutate.tiles)

    values = []
    for index in indexes:
        tile = widget.mutate.tiles[(index, 0)]
        tile_value_tile = tile.mutate.tiles[(0, 1)]
        tile_value_lines = tile_value_tile.mutate.lines
        tile_value = _helpers.join_lines(tile_value_lines)
        values.append(tile_value)

    values = (_helpers.paint_text(color, value) for value in values)

    value = delimiter.join(values)

    return value


@_theme.add('routines.basket')
def _basket(*args, 
            reply = _basket_reply, 
            **kwargs):
    
    """
    Use an :class:`~.widgets.Basket` widget.

    The default hint shows available controls.

    All widget and :func:`~.widgets.start` arguments are valid.

    |theme| :code:`'routines.basket'`

    .. image:: _static/images/routines.basket-1.gif

    .. code-block:: python

        days = ('Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday', 'Sunday')
        indexes = survey.routines.basket('Favourite days? ', options = days)
        print(f'Answered {indexes}.')

    .. image:: _static/images/routines.basket-2.gif

    .. code-block:: python

        names = tuple(map(''.join, itertools.combinations('AXBYCZ', 3)))
        def scout(spot):
            index = spot[0]
            name = names[index]
            return not 'C' in name
        index = survey.routines.basket('Favorite names? ',  options = names,  scout = scout,  evade_mark = '__',  focus_color = survey.colors.basic('magenta'), positive_mark = 'O ', negative_mark = 'X ',)
        print(f'Answered {index}.')
    """

    widget_kwargs_names = _helpers.get_function_args_names(_widgets.Basket)
    widget_kwargs = _helpers.yank_dict(kwargs, widget_kwargs_names)
    widget = _widgets.Basket(**widget_kwargs)

    kwargs.setdefault('hint', _basket_hint)
    kwargs.setdefault('hint_parse', True)

    def info(widget, *args):
        mutate = widget.mutate
        lines = mutate.search_lines
        point = mutate.search_point
        return (lines, point)

    multi_pre = True
    multi_aft = False

    result = _widgets.start(
        multi_pre, 
        multi_aft, 
        widget, 
        *args, 
        info = info,
        info_parse = False,
        site = 'info', 
        reply = reply, 
        **kwargs
    )

    return result


basket = _basket


@_theme.add('routines.datetime.reply')
def _datetime_reply(widget, 
                    result,
                    format = '%d/%m/%Y %H:%M',
                    color = _colors.basic('cyan')):

    result = result.strftime(format)
    
    result = _helpers.paint_text(color, result)

    return result


@_theme.add('routines.datetime')
def _datetime(*args,
             reply = _datetime_reply,
             **kwargs):
    
    """
    Use an :class:`~.widgets.DateTime` widget.

    All widget and :func:`~.widgets.start` arguments are valid.

    |theme| :code:`'routines.datetime'`

    .. image:: _static/images/routines.datetime-1.gif
    
    .. code-block:: python

        datetime = survey.routines.datetime('Schedule At: ')
        print(f'Answered {datetime}.')

    .. image:: _static/images/routines.datetime-2.gif
    
    .. code-block:: python

        datetime = survey.routines.datetime('Meeting Time: ', attrs = ('hour', 'minute'))
        print(f'Answered {datetime}.')
    """
    
    widget_kwargs_names = _helpers.get_function_args_names(_widgets.DateTime)
    widget_kwargs = _helpers.yank_dict(kwargs, widget_kwargs_names)
    widget = _widgets.DateTime(**widget_kwargs)

    multi_pre = False
    multi_aft = multi_pre

    result = _widgets.start(
        multi_pre, 
        multi_aft, 
        widget, 
        *args, 
        reply = reply, 
        **kwargs
    )

    return result


datetime = _datetime


_form_hint_gen_instructions_focus_extra = {
    _widgets.Select: _select_hint_gen_instructions,
    _widgets.Basket: _basket_hint_gen_instructions
}


def _form_hint_gen_instructions_focus(widget):

    sub_widget = widget.mutate.cur_tile.mutate.cur_tile

    try:
        extra = _form_hint_gen_instructions_focus_extra[sub_widget.__class__]
    except KeyError:
        pass
    else:
        yield from extra(sub_widget)

    yield 'back: submit'


def _form_hint_gen_instructions_blear(widget):

    yield from _select_hint_gen_instructions(widget)

    yield 'edit: <Tab>'


def _form_hint_gen_instructions(widget):

    if widget.focus:
        return _form_hint_gen_instructions_focus(widget)
    
    return _form_hint_gen_instructions_blear(widget)


_form_hint_gen_instructions_focus_extra[_widgets.Form] = _form_hint_gen_instructions


def _form_hint(widget, name, info):

    instructions = _form_hint_gen_instructions(widget)

    value = '[' + ' | '.join(dict.fromkeys(instructions)) + ']'

    return value


_form_reply_extra = {
    _widgets.Input: _input_reply,
    _widgets.Conceal: _input_reply,
    _widgets.Numeric: _numeric_reply,
    _widgets.Count: _numeric_reply,
    _widgets.Inquire: _inquire_reply,
    _widgets.Select: _select_reply,
    _widgets.Basket: _basket_reply,
    _widgets.DateTime: _datetime_reply,
}


@_theme.add('routines.form.reply')
def _form_reply(widget, 
                result,
                delimit = ': '):
    
    top_field_size = max(map(len, result))

    lines = []
    for widget in widget.mutate.tiles.values():
        field_widget = widget.mutate.tiles[(0, 0)]
        field = field_widget.resolve()
        value_widget = widget.mutate.tiles[(0, 1)]
        value = result[field]
        try:
            extra = _form_reply_extra[value_widget.__class__]
        except KeyError:
            value = str(value)
        else:
            value = extra(value_widget, value)
        lines.append(' ' * (top_field_size - len(field)) + field + delimit + value)

    result = '\n'.join(lines)

    return result


_form_reply_extra[_widgets.Form] = _form_reply


@_theme.add('routines.form')
def _form(*args,
          reply = _form_reply,
          **kwargs):
    
    """
    Use an :class:`~.widgets.Form` widget.

    The default hint shows available controls, including of focused widgets.

    All widget and :func:`~.widgets.start` arguments are valid.

    |theme| :code:`'routines.form'`

    .. image:: _static/images/routines.form-1.gif

    .. code-block:: python

        form = {
            'name': survey.widgets.Input(),
            'price': survey.widgets.Count(),
            'type': survey.widgets.Select(options = ('food', 'stationary', 'tobacco', 'literature'))
        }
        data = survey.routines.form('Item Data: ', form = form)
    """
    
    widget_kwargs_names = _helpers.get_function_args_names(_widgets.Form)
    widget_kwargs = _helpers.yank_dict(kwargs, widget_kwargs_names)
    widget = _widgets.Form(**widget_kwargs)

    kwargs.setdefault('hint', _form_hint)
    kwargs.setdefault('hint_parse', True)

    def info(widget, *args):
        mutate = widget.mutate
        lines = mutate.search_lines
        point = mutate.search_point
        return (lines, point)

    multi_pre = True
    multi_aft = True

    result = _widgets.start(
        multi_pre, 
        multi_aft, 
        widget, 
        *args, 
        reply = reply, 
        **kwargs
    )

    return result


form = _form




"""
Utilities for dealing with `ANSI Colors <https://en.wikipedia.org/wiki/ANSI_escape_code#Colors>`_.
"""

from . import _core
from . import _helpers


__all__ = ('style', 'basic', 'standard', 'full')


def _get(*args):

    return _core.ansi.get_control('m', *args)


_style_info = {
               'reset': 0,
              'strong': 1,
               'faint': 2,
              'italic': 3,
           'underline': 4,
          'slow_blink': 5,
         'rapid_blink': 6,
             'reverse': 7,
             'conceal': 8,
              'strike': 9,
    'underline_double': 21,
     'reset_intensity': 22,
        'reset_italic': 23,
     'reset_underline': 24,
         'reset_blink': 25,
       'reset_reverse': 27,
       'reset_conceal': 28,
        'reset_strike': 29,
      'reset_fg_color': 39,
      'reset_bg_color': 48
}


_style_info_doc = ', '.join(map('``{0}``'.format, _style_info))


def style(name: str) -> str:

    part = _style_info[name]

    value = _get(part)

    return _helpers.IdentifiableSGR(value, sl = name)

style.__doc__ = (
    """
    Get an ansi style by name.

    .. code-block:: python

        faint = colors.style('faint')

    The following styles are available: {0}
    """.format(_style_info_doc)
)


_color_bit4_info = {
      'black': ((30, 40), (90, 100)),
        'red': ((31, 41), (91, 101)),
      'green': ((32, 42), (92, 102)),
     'yellow': ((33, 43), (93, 103)),
       'blue': ((34, 44), (94, 104)),
    'magenta': ((35, 45), (95, 105)),
       'cyan': ((36, 46), (96, 106)),
      'white': ((37, 47), (97, 107)),
}


_color_bit4_light_indexes = {
    'dark': 0,
    'lite': 1
}


_color_bit4_layer_indexes = {
    'fg': 0,
    'bg': 1
}


def _color_bit4_resolve(path):

    (name, *rest) = path.split('.')

    info = {'name': name}

    part = _color_bit4_info[name]

    for (key, indexes) in (('light', _color_bit4_light_indexes), ('layer', _color_bit4_layer_indexes)):
        try:
            (name, *rest) = rest
        except ValueError:
            break
        info[key] = name
        index = indexes[name]
        part = part[index]

    return (part, info)


_color_default_light = 'lite'


def _color_bit4_get(fg_path, bg_path):

    default_light = _color_default_light

    store = []
    for (part, default_layer) in ((fg_path, 'fg'), (bg_path, 'bg')):
        if part is None:
            continue
        name, *rest = part.split('.')
        try:
            light, *rest = rest
        except ValueError:
            light = default_light
        light_index = _color_bit4_light_indexes[light]
        try:
            layer, *rest = rest
        except ValueError:
            layer = default_layer
        layer_index = _color_bit4_layer_indexes[layer]
        value = _color_bit4_info[name][light_index][layer_index]
        store.append(value)

    return _get(*store)


def basic(fg: str = None, bg: str = None) -> str:

    """
    Get a basic ansi color by path (`4 bit docs <https://en.wikipedia.org/wiki/ANSI_escape_code#3-bit_and_4-bit>`_). 

    :param fg:
        The foreground color.
    :param bg:
        The background color.

    .. code-block:: python

        fg_lite_blue = basic('blue.lite')
        fg_dark_green = basic('green.dark')
        bg_lite_yellow = basic(bg = 'yellow') # .lite can be ommited
        bg_dark_red_fg_dark_cyan = basic(fg = 'red.dark', bg = 'cyan.dark')
    """

    value = _color_bit4_get(fg, bg)
    
    return _helpers.IdentifiableSGR(value, fg = fg, bg = bg)


_color_default_layer = 'fg'


_color_bit8_info = {
    'fg': (38, 5),
    'bg': (48, 5)
}


def _color_bit8_info_get(layer, part):

    args = _color_bit8_info[layer]

    return _get(*args, part)


def standard(part: str, layer: str = _color_default_layer) -> str:

    """
    Get a standard ansi color by value (`8 bit docs <https://en.wikipedia.org/wiki/ANSI_escape_code#8-bit>`_).

    :param part:
        The color in ``[0, 255]`` range.
    :param layer:
        Whether it's foreground (:code:`'fg'`) or background (:code:`'bg'`). 

    .. code-block:: python

        fg_blue = standard(4)
        fg_pink = standard(225)
        bg_cyan = standard(6, 'bg')
    """
    
    return _color_bit8_info_get(layer, part)


_color_bit24_info = {
    'fg': (38, 2),
    'bg': (48, 2)
}


def _color_bit24_get(layer, part_r, part_g, part_b):

    args = _color_bit24_info[layer]

    return _get(*args, part_r, part_g, part_b)


def full(part_r: int, part_g: int, part_b: int, layer: str = _color_default_layer) -> str:

    """
    Get a full rgb ansi color by the respective values (`224 bit docs <https://en.wikipedia.org/wiki/ANSI_escape_code#24-bit>`_).

    :param part_r:
        The red component.
    :param part_g:
        The green component.
    :param part_b:
        The blue component.
    :parma layer:
        Whether it's foreground (:code:`'fg'`) or background (:code:`'bg'`). 

    .. code-block:: python

        fg_steel = full(113, 121, 126)
        bg_brick = fill(170, 74, 68, 'bg')
    """
    
    return _color_bit24_get(layer, part_r, part_g, part_b)